#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import os
from gi.repository import GLib

from soundconverter.util.fileoperations import unquote_filename, is_uri


class SoundFile:
    """Meta data information about a sound file (uri, tags)."""

    __slots__ = [
        'uri', 'base_path', 'filename', 'tags', 'filelist_row', 'subfolders',
        'readable', 'duration', 'info'
    ]

    def __init__(self, uri, base_path=None):
        """Create a SoundFile object.

        if base_path is set, the uri is cut in three parts,
         - the base folder, which is a common folder of multiple soundfiles
           (.base_path)
         - the remaining subfolders, which would be for example something
           like artist/album in the existing folder structure. As long as
           no subfolder pattern is provided, soundconverter will use those
           subfolders in the output directory. (in .subfolders)
         - the filename (in .filename)
        """
        # enforcing an uri format reduced the nightmare of handling 2
        # different path formats in generate_target_uri
        if not is_uri(uri):
            raise ValueError('uri was not an uri: {}!'.format(
                uri
            ))
        if base_path is not None and not is_uri(base_path):
            raise ValueError('base_path was not an uri: {}!'.format(
                base_path
            ))

        self.uri = uri
        self.subfolders = None

        if base_path:
            if not uri.startswith(base_path):
                raise ValueError(
                    'uri {} needs to start with the base_path {}!'.format(
                        uri, base_path
                    )
                )
            self.base_path = base_path
            subfolders, filename = os.path.split(uri[len(base_path):])
            self.subfolders = unquote_filename(subfolders)
            self.filename = filename
        else:
            self.base_path, self.filename = os.path.split(self.uri)
            self.base_path += '/'

        self.filelist_row = None

        # properties of valid audio are yet to be figured out in a Discoverer
        self.tags = {}
        self.readable = False
        self.duration = None
        self.info = None

    @property
    def filename_for_display(self):
        """Return the filename in a form suitable for displaying it."""
        return GLib.filename_display_name(
            unquote_filename(self.filename)
        )



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"""Batch mode to run soundconverter in a console."""

import os

from gi.repository import GLib, Gio

from soundconverter.interface.preferences import rates
from soundconverter.util.soundfile import SoundFile
from soundconverter.util.settings import settings, set_gio_settings
from soundconverter.util.formats import get_quality_setting_name, \
    get_mime_type, get_mime_type_mapping, get_default_quality
from soundconverter.gstreamer.converter import Converter
from soundconverter.gstreamer.discoverer import add_discoverers, \
    get_sound_files
from soundconverter.util.taskqueue import TaskQueue
from soundconverter.util.namegenerator import TargetNameGenerator
from soundconverter.util.fileoperations import filename_to_uri, beautify_uri
from soundconverter.util.logger import logger
from soundconverter.util.formatting import format_time
from soundconverter.interface.mainloop import gtk_iteration

cli_convert = [None]


def batch_main(files):
    global cli_convert
    # works like a pointer, so that it can be accessed in tests,
    # just like in ui.py
    cli_convert[0] = CLIConvert(files)


def use_memory_gsettings(options):
    """Use a Gio memory backend and write argv settings into it.

    In order for the batch mode to work properly with functions that were
    written for the ui, write argv into the gio settings, but provide a
    temporary memory backend so that the ui keeps its settings and cli
    settings are thrown away at the end.
    """
    if options.get('main') == 'gui':
        raise AssertionError('should not be used with the gui mode')

    backend = Gio.memory_settings_backend_new()
    gio_settings = Gio.Settings.new_with_backend('org.soundconverter', backend)
    set_gio_settings(gio_settings)

    gio_settings.set_boolean('delete-original', options.get('delete-original', False))

    if options.get('main') == 'batch':
        # the number of jobs is only applied, when limit-jobs is true
        forced_jobs = options.get('forced-jobs', None)
        if forced_jobs is not None:
            gio_settings.set_boolean('limit-jobs', True)
            gio_settings.set_int('number-of-jobs', options['forced-jobs'])
        else:
            gio_settings.set_boolean('limit-jobs', False)

        format_option = options['format']
        mime_type = get_mime_type(format_option)

        mode = options.get('mode')
        if mode:
            gio_settings.set_string('mp3-mode', mode)

        gio_settings.set_string('output-mime-type', mime_type)

        output_path = filename_to_uri(options.get('output-path'))
        gio_settings.set_string('selected-folder', output_path)

        gio_settings.set_boolean('same-folder-as-input', False)

        # other than the ui, the batch mode doesn't have a selection of
        # predefined patterns and it's all text anyways, so patterns can be
        # created with the -p argument on custom-filename-pattern. Don't use
        # -o for that to avoid having to escape pattern symbols when a folder
        # is called like a pattern string.
        gio_settings.set_boolean('create-subfolders', False)
        pattern = options.get('custom-filename-pattern')
        if pattern:
            gio_settings.set_int('name-pattern-index', -1)
            gio_settings.set_string('custom-filename-pattern', pattern)
        else:
            # the first name pattern is the filename itself
            gio_settings.set_int('name-pattern-index', 0)

        quality_setting = options.get('quality')
        if quality_setting is None:
            quality_setting = get_default_quality(mime_type)

        setting_name = get_quality_setting_name()
        # here is the very long and incredible way to set a variable as gio
        # settings value with the correct type as defined in the schema:
        type_string = gio_settings \
            .props \
            .settings_schema \
            .get_key(setting_name) \
            .get_value_type() \
            .dup_string()
        variant = GLib.Variant(type_string, float(quality_setting))
        gio_settings.set_value(
            setting_name,
            variant
        )

    else:
        # --tags and --check
        gio_settings.set_boolean('limit-jobs', True)
        gio_settings.set_int('number-of-jobs', 1)

    resample = options.get('output-resample')

    if resample is not None:
        gio_settings.set_boolean('output-resample', True)
        gio_settings.set_int('resample-rate', resample)


def validate_args(options):
    """Check if required command line args are provided.

    Will log usage mistakes to the console.
    """
    main = options.get('main', 'gui')
    if main not in ['gui', 'check', 'tags', 'batch']:
        logger.error('unknown main {}'.format(main))
        return False

    if main not in ['gui', 'check', 'tags']:
        # not needed for --check and --tags
        if not options.get('output-path'):
            logger.error('output path argument -o is required')
            return False

        existing_behaviour = options.get('existing')
        if existing_behaviour:
            existing_behaviours = [
                Converter.SKIP, Converter.OVERWRITE, Converter.INCREMENT
            ]
            if existing_behaviour not in existing_behaviours:
                logger.error('-e should be one of {}'.format(
                    ', '.join(existing_behaviours)
                ))
                return False

        # target_format might be a mime type or a file extension
        target_format = options.get('format')
        if target_format is None:
            logger.error('format argument -f is required')
            return False
        mime_type = get_mime_type(target_format)
        if mime_type is None:
            logger.error(
                'cannot use "{}" format. Supported formats: {}'.format(
                    target_format,
                    ', '.join(get_mime_type_mapping())
                )
            )
            return False

        mode = options.get('mode')
        if mode and mode not in ['abr', 'cbr', 'vbr']:
            logger.error('mode should be one of abr, cbr or vbr (default)')
            return False

        # validate if the quality setting makes sense
        quality = options.get('quality')

        if quality is not None:
            # optional; otherwise default quality values will be used
            if mime_type == 'audio/mpeg':
                if mode in ['abr', 'cbr']:
                    if quality > 320 or quality < 64:
                        logger.error(
                            'mp3 cbr/abr bitrate should be between 64 and 320'
                        )
                        return False
                else:
                    if quality > 9 or quality < 0:
                        logger.error(
                            'mp3 vbr quality should be between 9 (low) and '
                            '0 (hight)'
                        )
                        return False

            elif mime_type == 'audio/x-vorbis':
                if quality < 0 or quality > 1:
                    logger.error('ogg quality should be between 0.0 and 1.0')
                    return False

            elif mime_type == 'audio/x-m4a':
                # supports arbitrary bitrates
                # source: https://en.wikipedia.org/wiki/Advanced_Audio_Coding
                # Our used encoder seems to cap somewhere between 400 and
                # 440 kbps though
                if quality < 0:
                    logger.error('m4a bitrate should be larger than 0')
                    return False

            elif mime_type == 'audio/x-flac':
                if quality < 0 or quality > 8:
                    logger.error(
                        'flac compression strength should be between 0 and 8'
                    )
                    return False

            elif mime_type == 'audio/x-wav':
                if quality not in [8, 16, 24, 32]:
                    logger.error(
                        'wav sample width has to be one of 8, 16, 24 or 32'
                    )
                    return False

            elif mime_type == 'audio/ogg; codecs=opus':
                # source: https://wiki.hydrogenaud.io/index.php?title=Opus
                if quality < 6 or quality > 510:
                    logger.error('opus bitrate should be between 6 and 510')
                    return False

    resample = options.get('output-resample', None)

    if resample and resample not in rates:
        logger.error(
            "Possible values for -R or --output-resample option are {rates}".format(
                rates=", ".join([str(item for item in rates)])
            )
        )
        return False

    return True


def prepare_files_list(input_files):
    """Create a list of all URIs in a list of paths.

    Also returns a list of relative directories. This is used to reconstruct
    the directory structure in the output path.

    If input_path is a/b/c/ and the file is at a/b/c/e/f/d.mp3,
    the subdirectries entry will be subdirectories c/e/f/.

    If input_files is ['/a/b', '/c']
    and files are found at ['file:///a/b/d.mp3', '[file:///c/e/f/g.mp3'],
    subdirectories will be ['b/d', 'c/e/f']

    Subdirectories might be different for various files because
    multiple paths can be provided in the args.

    Parameters
    ----------
    input_files : string[]
        Array of paths (not uris)
    """
    # The GUI has its own way of going through subdirectories.

    # If one of the files is a directory, walk over the files in that
    # and append each one to parsed_files if -r is provided.
    subdirectories = []
    parsed_files = []
    for input_path in input_files:
        # accept tilde (~) to point to home directories, get absolute path
        input_path = os.path.realpath(os.path.expanduser(input_path))

        if os.path.isfile(input_path):
            # for every appended file, also append to
            # subdirectories
            subdirectories.append('')
            parsed_files.append(input_path)
        elif not os.path.isdir(input_path):
            logger.error('path {} does not exist'.format(input_path))

        # walk over directories to add the files of all the subdirectories
        elif os.path.isdir(input_path):
            if input_path[-1] == os.sep:
                input_path = input_path[:-1]

            if input_path.rfind(os.sep) != -1:
                parent = input_path[:input_path.rfind(os.sep)]
            else:
                parent = input_path

            # but only if -r option was provided
            if settings.get('recursive'):
                for dirpath, _, filenames in os.walk(input_path):
                    for filename in filenames:
                        if dirpath[-1] != os.sep:
                            dirpath += os.sep
                        # dirpath: for example a/b/c/e/f/
                        parsed_files.append(dirpath + filename)
                        subdir = os.path.relpath(dirpath, parent) + os.sep
                        if subdir == './':
                            subdir = ''
                        subdirectories.append(subdir)
            else:
                # else it didn't go into any directory.
                # Provide some information about how to
                logger.error(
                    '{} is a directory. Use -r to go into all subdirectories.'
                    .format(input_path)
                )
        # if not a file and not a dir it doesn't exist. skip

    parsed_files = list(map(filename_to_uri, parsed_files))

    return parsed_files, subdirectories


class CLIConvert:
    """Main class that runs the conversion."""

    def __init__(self, input_files):
        """Start the conversion of all the files specified in input_files.

        To control the conversion and the handling of directories, command
        line arguments have to be provided which are stored in the global
        'settings' variable.

        input_files is an array of string paths.
        """
        logger.info(
            'checking files and walking dirs in the specified paths…'
        )

        # CLICheck will exit(1) if no input_files available and resolve
        # all files in input_files and also figure out which files can be
        # converted.
        file_checker = CLICheck(input_files)

        loop = GLib.MainLoop()
        context = loop.get_context()

        name_generator = TargetNameGenerator()
        suffix = name_generator.suffix
        name_generator.suffix = suffix

        conversions = TaskQueue()

        self.started_tasks = 0
        self.num_conversions = 0

        sound_files = file_checker.get_sound_files()
        for sound_file in sound_files:
            if not sound_file.readable:
                filename = beautify_uri(sound_file.uri)
                logger.info(
                    'skipping \'{}\': not an audiofile'.format(filename)
                )
                continue

            converter = Converter(sound_file, name_generator)

            converter.existing_behaviour = settings.get('existing')

            conversions.add(converter)

            self.num_conversions += 1

        if self.num_conversions == 0:
            logger.info('no audio files for conversion found…')
            exit(2)

        logger.info('starting conversion of {} files…'.format(
            len(sound_files)
        ))
        self.conversions = conversions

        def finished(_=None):
            total_time = conversions.get_duration()
            logger.info('converted {} files in {}'.format(
                len(sound_files),
                format_time(total_time)
            ))

        conversions.connect('done', finished)

        conversions.run()

        while not conversions.finished:
            # make the eventloop of glibs async stuff run until finished:
            context.iteration(may_block=True)

        # do another one to print the queue done message
        context.iteration(may_block=False)


class CLICheck:
    """Print all the tags of the specified files to the console."""

    def __init__(
            self, input_files, verbose=False
    ):
        """Print all the tags of the specified files to the console.

        To go into subdirectories of paths provided, the -r command line
        argument should be provided, which is stored in the global 'settings'
        variable.

        It will exit soundconverter if input_files contains no files
        (maybe because -r is missing and the specified path is a dir)

        Parameters
        ----------
        input_files : string[]
            an array of string paths.
        verbose : bool
            if True, will print tags, readable and non readable paths
        """

        input_files, subdirectories = prepare_files_list(input_files)

        if len(input_files) == 0:
            # prepare_files_list will print something like
            # "use -r to go into subdirectories"
            logger.info('no files found…')
            exit(1)

        discoverers = TaskQueue()
        sound_files = []
        for subdirectory, input_file in zip(subdirectories, input_files):
            sound_file = SoundFile(input_file)
            # by storing it in subfolders, the original subfolder structure
            # (relative to the directory that was provided as input in the
            # cli) can be restored in the target dir
            sound_file.subfolders = subdirectory
            sound_files.append(sound_file)

        add_discoverers(discoverers, sound_files)
        discoverers.run()

        loop = GLib.MainLoop()
        context = loop.get_context()

        while not discoverers.finished:
            # main_iteration is needed for the typefinder taskqueue to run
            # calling this like crazy is the fastest way
            context.iteration(True)

        self.discoverers = discoverers.all_tasks
        sound_files = []
        for discoverer in self.discoverers:
            sound_files += discoverer.sound_files

        if verbose:
            for sound_file in sound_files:
                if sound_file.readable:
                    self.print(sound_file)
                else:
                    logger.info('{} is not an audiofile'.format(
                        beautify_uri(sound_file.uri)
                    ))

    def get_sound_files(self):
        """Get all SoundFiles."""
        return get_sound_files(self.discoverers)

    def print(self, sound_file):
        """Print tags of a file, or write that it doesn't have tags."""
        logger.info(beautify_uri(sound_file.uri))

        if len(sound_file.tags) > 0:
            for key in sorted(sound_file.tags):
                value = sound_file.tags[key]
                logger.info(('    {}: {}'.format(key, value)))
        else:
            logger.info(('    no tags found'))



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA


import os
from gettext import gettext as _

from gi.repository import Gst, GLib, Gio

from soundconverter.util.fileoperations import vfs_encode_filename, \
    vfs_unlink, vfs_rename, vfs_exists, beautify_uri
from soundconverter.util.logger import logger
from soundconverter.util.settings import get_gio_settings
from soundconverter.util.error import show_error
from soundconverter.util.task import Task


GSTREAMER_SOURCE = 'giosrc'
GSTREAMER_SINK = 'giosink'

available_elements = set()


def find_available_elements():
    """Figure out which gstreamer pipeline plugins are available."""
    # gst-plugins-good, gst-plugins-bad, etc. packages provide them
    global available_elements

    GOOD = 'gst-plugins-good'
    BAD = 'gst-plugins-bad'
    BASE = 'gst-plugins-base'
    UGLY = 'gst-plugins-ugly'
    LIBAV = 'gst-libav'

    # some functions can be provided by various packages.
    # move preferred packages towards the bottom.
    encoders = [
        ('flacenc', 'FLAC', 'flac-enc', GOOD),
        ('wavenc', 'WAV', 'wav-enc', GOOD),
        ('vorbisenc', 'Ogg Vorbis', 'vorbis-enc', BASE),
        ('oggmux', 'Ogg Vorbis', 'vorbis-mux', BASE),
        ('id3mux', 'MP3 tags', 'mp3-id-tags', BAD),
        ('id3v2mux', 'MP3 tags', 'mp3-id-tags', GOOD),
        ('xingmux', 'VBR tags', 'mp3-vbr-tags', UGLY),
        ('lamemp3enc', 'MP3', 'mp3-enc', GOOD),
        ('mp4mux', 'AAC', 'aac-mux', GOOD),
        ('opusenc', 'Opus', 'opus-enc', BASE),
        ('faac', 'AAC', 'aac-enc', BAD),
        ('avenc_aac', 'AAC', 'aac-enc', LIBAV),
        ('fdkaacenc', 'AAC', 'aac-enc', BAD),
    ]

    result = dict()
    for encoder, name, function, package in encoders:
        have_it = bool(Gst.ElementFactory.find(encoder))
        if have_it:
            available_elements.add(encoder)
        else:
            logger.debug(
                '{} gstreamer element from "{}" not found\n'.format(
                    encoder,
                    package
                )
            )
        result[function] = (have_it, package)

    for function in result:
        have_it, package = result[function]
        if not have_it:
            logger.error(
                'Disabling {} output. Do you have "{}" installed?'.format(
                    function,
                    package
                )
            )

    if 'oggmux' not in available_elements:
        available_elements.discard('vorbisenc')
    if 'mp4mux' not in available_elements:
        available_elements.discard('faac')
        available_elements.discard('avenc_aac')


find_available_elements()


def create_flac_encoder():
    """Return an flac encoder for the gst pipeline string."""
    flac_compression = get_gio_settings().get_int('flac-compression')
    return 'flacenc mid-side-stereo=true quality={}'.format(flac_compression)


def create_wav_encoder():
    """Return a wav encoder for the gst pipeline string."""
    wav_sample_width = get_gio_settings().get_int('wav-sample-width')
    formats = {8: 'U8', 16: 'S16LE', 24: 'S24LE', 32: 'S32LE'}
    return 'audioconvert ! audio/x-raw,format={} ! wavenc'.format(
        formats[wav_sample_width]
    )


def create_oggvorbis_encoder():
    """Return an ogg encoder for the gst pipeline string."""
    cmd = 'vorbisenc'
    vorbis_quality = get_gio_settings().get_double('vorbis-quality')
    if vorbis_quality is not None:
        cmd += ' quality={}'.format(vorbis_quality)
    cmd += ' ! oggmux '
    return cmd


def create_mp3_encoder():
    """Return an mp3 encoder for the gst pipeline string."""
    quality = {
        'cbr': 'mp3-cbr-quality',
        'abr': 'mp3-abr-quality',
        'vbr': 'mp3-vbr-quality'
    }
    mode = get_gio_settings().get_string('mp3-mode')

    mp3_mode = mode
    mp3_quality = get_gio_settings().get_int(quality[mode])

    cmd = 'lamemp3enc encoding-engine-quality=2 '

    if mp3_mode is not None:
        properties = {
            'cbr': 'target=bitrate cbr=true bitrate=%s ',
            'abr': 'target=bitrate cbr=false bitrate=%s ',
            'vbr': 'target=quality cbr=false quality=%s ',
        }

        cmd += properties[mp3_mode] % mp3_quality

        if 'xingmux' in available_elements and mp3_mode != 'cbr':
            # add xing header when creating vbr/abr mp3
            cmd += '! xingmux '

    if 'id3mux' in available_elements:
        # add tags
        cmd += '! id3mux '
    elif 'id3v2mux' in available_elements:
        # add tags
        cmd += '! id3v2mux '

    return cmd


def create_aac_encoder():
    """Return an aac encoder for the gst pipeline string."""
    aac_quality = get_gio_settings().get_int('aac-quality')

    # it seemed like I couldn't get vbr to work with any of these, not even
    # with rate-control and quality of faac or with maxrate of avenc_aac.
    # Or it was audacious not displaying the current vbr rate correctly.
    bitrate = aac_quality * 1000

    # list of recommended aac encoders:
    # https://wiki.hydrogenaud.io/index.php?title=AAC_encoders
    if 'fdkaacenc' in available_elements:
        return 'fdkaacenc bitrate={} ! mp4mux'.format(bitrate)

    encoder = 'faac' if 'faac' in available_elements else 'avenc_aac'
    logger.warning(
        'fdkaacenc is recommended for aac conversion but it is not '
        'available. It can be installed with gst-plugins-bad. '
        'Using {} instead.'.format(encoder)
    )

    if 'faac' in available_elements:
        return 'faac bitrate={} rate-control=2 ! mp4mux'.format(bitrate)

    return 'avenc_aac bitrate={} ! mp4mux'.format(bitrate)


def create_opus_encoder():
    """Return an opus encoder for the gst pipeline string."""
    opus_quality = get_gio_settings().get_int('opus-bitrate')
    return (
        'opusenc bitrate={} bitrate-type=vbr '
        'bandwidth=auto ! oggmux'
    ).format(opus_quality * 1000)


class Converter(Task):
    """Completely handle the conversion of a single file."""
    INCREMENT = 'increment'
    OVERWRITE = 'overwrite'
    SKIP = 'skip'

    def __init__(self, sound_file, name_generator):
        """create a converter that converts a single file.

        Parameters
        ----------
        name_generator : TargetNameGenerator
            TargetNameGenerator that creates filenames for all converters
            of the current TaskQueue
        """
        # Configuration
        self.sound_file = sound_file
        self.temporary_filename = None
        self.newname = None
        self.existing_behaviour = Converter.INCREMENT
        self.name_generator = name_generator

        # All relevant gio settings have to be copied and remembered, so that
        # they don't suddenly change during the conversion
        settings = get_gio_settings()
        self.output_mime_type = settings.get_string('output-mime-type')
        self.output_resample = settings.get_boolean('output-resample')
        self.resample_rate = settings.get_int('resample-rate')
        self.force_mono = settings.get_boolean('force-mono')
        self.replace_messy_chars = settings.get_boolean('replace-messy-chars')
        self.delete_original = settings.get_boolean('delete-original')

        # State
        self.command = None
        self.pipeline = None
        self._done = False
        self.error = None
        self.output_uri = None

        super().__init__()

    def _query_position(self):
        """Ask for the stream position of the current pipeline."""
        if self.pipeline:
            # during Gst.State.PAUSED it returns super small numbers,
            # so take care
            position = self.pipeline.query_position(Gst.Format.TIME)[1]
            return max(0, position / Gst.SECOND)
        return 0

    def get_progress(self):
        """Fraction of how much of the task is completed."""
        duration = self.sound_file.duration

        if self._done:
            return 1, duration

        if self.pipeline is None or duration is None:
            return 0, duration

        position = self._query_position()
        progress = position / duration if duration else 0
        progress = min(max(progress, 0.0), 1.0)
        return progress, duration

    def cancel(self):
        """Cancel execution of the task."""
        self._stop_pipeline()
        self.done()

    def pause(self):
        """Pause execution of the task."""
        if not self.pipeline:
            logger.debug('pause(): pipeline is None!')
            return
        self.pipeline.set_state(Gst.State.PAUSED)

    def resume(self):
        """Resume execution of the task."""
        if not self.pipeline:
            logger.debug('resume(): pipeline is None!')
            return
        self.pipeline.set_state(Gst.State.PLAYING)

    def _cleanup(self):
        """Delete the pipeline."""
        if self.pipeline is not None:
            bus = self.pipeline.get_bus()
            if hasattr(self, 'watch_id'):
                bus.disconnect(self.watch_id)
                bus.remove_signal_watch()
            self.pipeline.set_state(Gst.State.NULL)
            self.pipeline = None

    def _stop_pipeline(self):
        # remove partial file
        if self.temporary_filename is not None:
            if vfs_exists(self.temporary_filename):
                try:
                    vfs_unlink(self.temporary_filename)
                except Exception as error:
                    logger.error('cannot delete: \'{}\': {}'.format(
                        beautify_uri(self.temporary_filename),
                        str(error)
                    ))
        if not self.pipeline:
            logger.debug('pipeline already stopped!')
            return
        self._cleanup()

    def _convert(self):
        """Run the gst pipeline that converts files.

        Handlers for messages sent from gst are added, which also triggers
        renaming the file to it's final path.
        """
        command = self.command
        if self.pipeline is None:
            logger.debug('launching: \'{}\''.format(command))
            try:
                self.pipeline = Gst.parse_launch(command)
                bus = self.pipeline.get_bus()
            except GLib.Error as error:
                self.error = 'gstreamer error when creating pipeline: {}'.format(str(error))
                self._on_error(self.error)
                return

            bus.add_signal_watch()
            self.watch_id = bus.connect('message', self._on_message)

        self.pipeline.set_state(Gst.State.PLAYING)

    def _conversion_done(self):
        """Should be called when the EOS message arrived or on error.

        Will clear the temporary data on error or move the temporary file
        to the final path on success.
        """
        input_uri = self.sound_file.uri
        newname = self.newname

        if newname is None:
            raise AssertionError('the conversion was not started')

        if self.error:
            logger.debug('error in task, skipping rename: {}'.format(
                self.temporary_filename
            ))
            vfs_unlink(self.temporary_filename)
            logger.error('could not convert {}: {}'.format(
                beautify_uri(input_uri),
                self.error
            ))
            self.done()
            return

        if not vfs_exists(self.temporary_filename):
            self.error = 'Expected {} to exist after conversion.'.format(
                self.temporary_filename
            )
            self.done()
            return

        # rename temporary file
        logger.debug('{} -> {}'.format(
            beautify_uri(self.temporary_filename), beautify_uri(newname)
        ))

        path, extension = os.path.splitext(newname)
        path = path.replace('%', '%%')

        space = ' '
        if self.replace_messy_chars:
            space = '_'

        exists = vfs_exists(newname)
        if self.existing_behaviour == Converter.INCREMENT and exists:
            # If the file already exists, increment the filename so that
            # nothing gets overwritten.
            path = path + space + '(%d)' + extension
            i = 1
            while vfs_exists(newname):
                newname = path % i
                i += 1

        try:
            if self.existing_behaviour == Converter.OVERWRITE and exists:
                logger.info('overwriting \'{}\''.format(
                    beautify_uri(newname)
                ))
                vfs_unlink(newname)
            vfs_rename(self.temporary_filename, newname)
        except Exception as error:
            self.error = str(error)
            logger.error("could not rename '{}' to '{}':".format(
                beautify_uri(self.temporary_filename),
                beautify_uri(newname),
                str(error)
            ))
            self.done()
            return

        assert vfs_exists(newname)

        logger.info("converted '{}' to '{}'".format(
            beautify_uri(input_uri), beautify_uri(newname)
        ))

        # finish up the target file
        try:
            # Copy file permissions
            source = Gio.file_parse_name(self.sound_file.uri)
            destination = Gio.file_parse_name(newname)
            source.copy_attributes(destination, Gio.FileCopyFlags.NONE)

            # the modification date of the destination should be now
            info = Gio.FileInfo()
            now = GLib.DateTime.new_now(GLib.TimeZone())
            if callable(getattr(info, 'set_modification_date_time', None)):
                info.set_modification_date_time(now)
            else:
                # deprecated method
                timeval = GLib.TimeVal()
                now.to_timeval(timeval)
                info.set_modification_time(timeval)

            destination.set_attributes_from_info(
                info,
                Gio.FileQueryInfoFlags.NONE
            )
        except Exception as error:
            logger.error(
                "Could not set some attributes of the target '{}': {}".format(
                    beautify_uri(newname),
                    str(error)
                )
            )

        if self.delete_original and not self.error:
            logger.info("deleting: '{}'".format(self.sound_file.uri))
            try:
                vfs_unlink(self.sound_file.uri)
            except Exception as error:
                logger.info("cannot remove '{}': {}".format(
                    beautify_uri(self.sound_file.uri),
                    str(error)
                ))

        self.output_uri = newname
        self.done()

    def done(self):
        self._done = True
        self._cleanup()
        super().done()

    def run(self):
        """Call this in order to run the whole Converter task."""
        self.newname = self.name_generator.generate_target_uri(
            self.sound_file
        )

        # temporary output file, in order to easily remove it without
        # any overwritten file and therefore caused damage in the target dir.
        self.temporary_filename = self.name_generator.generate_temp_path(
            self.sound_file
        )

        exists = vfs_exists(self.newname)
        if self.existing_behaviour == Converter.SKIP and exists:
            logger.info('output file already exists, skipping \'{}\''.format(
                beautify_uri(self.newname)
            ))
            self.done()
            return

        # construct a pipeline for conversion
        # Add default decoding step that remains the same for all formats.
        command = ['{} location="{}" name=src ! decodebin name=decoder'.format(
            GSTREAMER_SOURCE, vfs_encode_filename(self.sound_file.uri)
        ), 'audiorate ! audioconvert ! audioresample']

        # audio resampling support
        if self.output_resample:
            command.append('audio/x-raw,rate={}'.format(self.resample_rate))
            command.append('audioconvert ! audioresample')

        if self.force_mono:
            command.append('audio/x-raw,channels=1 ! audioconvert')

        # figure out the rest of the gst pipeline string
        encoder = {
            'audio/x-vorbis': create_oggvorbis_encoder,
            'audio/x-flac': create_flac_encoder,
            'audio/x-wav': create_wav_encoder,
            'audio/mpeg': create_mp3_encoder,
            'audio/x-m4a': create_aac_encoder,
            'audio/ogg; codecs=opus': create_opus_encoder
        }[self.output_mime_type]()
        command.append(encoder)

        gfile = Gio.file_parse_name(self.temporary_filename)
        dirname = gfile.get_parent()
        if dirname and not dirname.query_exists(None):
            logger.info('creating folder: \'{}\''.format(
                beautify_uri(dirname.get_uri())
            ))
            if not dirname.make_directory_with_parents():
                show_error(
                    _('cannot create \'{}\' folder.').format(
                        beautify_uri(dirname)
                    )
                )
                return

        command.append('{} location="{}"'.format(
            GSTREAMER_SINK, vfs_encode_filename(self.temporary_filename)
        ))

        # preparation done, now convert
        self.command = ' ! '.join(command)
        self._convert()

    def _on_error(self, error):
        """Log errors and write down that this Task failed.

        The TaskQueue is interested in reading the error.
        """
        self.error = error
        show_error(
            error,
            beautify_uri(self.sound_file.uri)
        )
        self._stop_pipeline()
        self.done()

    def _on_message(self, _, message):
        """Handle message events sent by gstreamer.

        Parameters
        ----------
        message : Gst.Message
        """
        if message.type == Gst.MessageType.ERROR:
            error, __ = message.parse_error()
            self._on_error(error)
        elif message.type == Gst.MessageType.EOS:
            # Conversion done
            self._conversion_done()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import time
from queue import Queue

from gi.repository import GObject, GLib

from soundconverter.util.settings import get_num_jobs
from soundconverter.interface.mainloop import gtk_iteration


class TaskQueue(GObject.Object):
    """Executes multiple tasks in parallel."""
    def __init__(self):
        self._on_queue_finished = None

        # state
        self.all_tasks = []
        self.pending = Queue()
        self.running = []
        self.done = []
        self.finished = False
        self.paused = False
        self._timer = Timer()

        super().__init__()

    def add(self, task):
        """Add a task to the queue that will be executed later.

        Parameters
        ----------
        task : Task
            Any object inheriting from Task
        """
        task.timer = Timer()
        self.all_tasks.append(task)
        self.pending.put(task)

    def get_progress(self, only_running=False):
        """Get the fraction of tasks that have been completed.

        returns a tuple of (total progress, task progress)
        with "task progress" being a list of (task, progress) tuples.
        """
        # some tasks may take longer, in order to communicate that they
        # provide a weight attribute.
        if len(self.all_tasks) == 0:
            return None
        total_weight = 0
        total_progress = 0
        tasks = self.running if only_running else self.all_tasks

        task_progress = []

        for task in tasks:
            progress, weight = task.get_progress()
            total_progress += progress * weight
            total_weight += weight
            task_progress.append((task, progress))

        return total_progress / total_weight, task_progress

    def pause(self):
        """Pause all tasks."""
        self._timer.pause()
        self.paused = True
        for task in self.running:
            task.timer.pause()
            task.pause()

    def resume(self):
        """Resume all tasks after the queue has been paused."""
        self._timer.resume()
        for task in self.running:
            task.timer.resume()
            task.resume()
        self.paused = False

    def cancel(self):
        """Stop all tasks."""
        self.finished = True
        for task in self.running:
            # by calling run it can be resumed, but cancelled tasks will start
            # from the beginning. The proper way would be to call pause and
            # resume for such a functionality though.
            self.pending.put(task)
            task.timer.stop()
            task.cancel()
        self._timer.reset()
        self.running = []

    def task_done(self, task):
        """One task is done, start another one.

        This callback has to be called by the task, when the task is done.

        Parameters
        ----------
        task : Task
            A completed task
        """
        # avoid adding duplicate signal handlers if the queue is restarted
        task.disconnect_by_func(self.task_done)

        if task in self.done:
            raise Exception('Duplicate task_done call')

        if self.finished:
            return

        task.timer.stop()

        self.done.append(task)
        self.running.remove(task)

        if self.pending.qsize() > 0:
            self.start_next()
        elif len(self.running) == 0:
            self.finished = True
            self._timer.stop()
            self.emit('done')

    def start_next(self, _=None):
        """Start the next task if available."""
        if self.pending.qsize() > 0:
            task = self.pending.get()

            task.connect('done', self.task_done)

            self.running.append(task)

            # - Just looping over self.pending causes too many tasks to be running in
            # parallel
            # - There is no semaphore mechanism for glib to do it in a simple while
            # loop, in order to limit it to num_jobs
            # - I don't think GLib.Mutex works with main-loop single-thread
            # parallelization and the whole thread will stop here
            # - Telling tasks to start the next task via a python callback or event
            # causes the runtime to crash due to recursion with too many tasks.
            # However, idle_add prevents those recursion depth problems
            GLib.idle_add(task.run)
            task.timer.start()

    def run(self):
        """Run all tasks."""
        self.finished = False
        self._timer.start()
        num_jobs = get_num_jobs()

        while self.pending.qsize() > 0 and len(self.running) < num_jobs:
            # Run as many tasks as the configured number of jobs. Finished tasks will
            # trigger running the next task via a event
            self.start_next()

        gtk_iteration()

    def get_duration(self):
        """Get for how many seconds the queue has been actively running.

        The time spent while being paused is not included.
        """
        return self._timer.get_duration()

    def get_remaining(self):
        """Calculate how many seconds are left until the queue is done."""
        if len(self.running) == 0:
            # cannot be estimated yet
            return None

        total_duration = 0
        total_remaining_weight = 0
        total_processed_weight = 0

        max_remaining_weight = -1
        for task in self.all_tasks:
            # duration is the time the timer has been running, not the
            # audio duration.
            duration = task.timer.get_duration()
            # total_duration would be 12s if 12 tasks run for 1s
            total_duration += duration

            # weight is actually the audio duration, but it's unit is going
            # to be canceled in the remaining_duration calculation. It could
            # be anything as long as all tasks have the same unit of weight.
            progress, weight = task.get_progress()
            remaining_weight = (1 - progress) * weight
            max_remaining_weight = max(remaining_weight, max_remaining_weight)
            total_remaining_weight += remaining_weight
            processed_weight = progress * weight
            total_processed_weight += processed_weight

        if total_processed_weight == 0:
            # cannot be calculated yet
            return None

        # how many seconds per weight. This remains pretty stable, even when
        # less processes are running in parallel, because total_duration
        # is the sum of all task durations and not the queues duration.
        speed = total_duration / total_processed_weight

        # how much weight left per process
        remaining_weight_per_p = total_remaining_weight / len(self.running)
        remaining_duration = speed * remaining_weight_per_p

        # if the max_remaining time exceeds the time of the
        # remaining_duration which especially happens when the conversion
        # comes to an end while one very large file is being converted,
        # take that one.
        if max_remaining_weight != -1:
            max_remaining = speed * max_remaining_weight
            remaining = max(max_remaining, remaining_duration)
        else:
            remaining = remaining_duration

        return remaining


GObject.signal_new(
    'done',
    TaskQueue,
    GObject.SignalFlags.RUN_FIRST,
    None,
    []
)


class Timer:
    """Time how long the TaskQueue took."""
    # separate class because I would like to not pollute the TaskQueue
    # with a bunch of timing variables
    def __init__(self):
        self.reset()

    def reset(self):
        self.run_start_time = None
        self.pause_duration = 0
        self.pause_time = None
        self.finished_time = None

    def stop(self):
        self.finished_time = time.time()

    def start(self):
        self.run_start_time = time.time()

    def pause(self):
        self.pause_time = time.time()

    def resume(self):
        self.pause_duration += time.time() - self.pause_time
        self.pause_time = None

    def get_duration(self):
        """Get for how many seconds the queue has been actively running.

        The time spent while being paused is not included.
        """
        if self.run_start_time is None:
            return 0
        if self.pause_time is not None:
            # still being paused
            pause_duration = time.time() - self.pause_time
        else:
            pause_duration = self.pause_duration
        if self.finished_time is None:
            # still running
            finished_time = time.time()
        else:
            finished_time = self.finished_time
        return finished_time - self.run_start_time - pause_duration



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

# logging & debugging

from soundconverter.util.settings import settings
import logging


class Formatter(logging.Formatter):
    def format(self, record):
        if record.levelno == logging.INFO and not settings['debug']:
            # if not launched with --debug, then don't print "INFO:"
            self._style._fmt = '%(msg)s'  # noqa
        else:
            # see https://en.wikipedia.org/wiki/ANSI_escape_code#3/4_bit
            # for those numbers
            color = {
                logging.WARNING: 33,
                logging.ERROR: 31,
                logging.FATAL: 31,
                logging.DEBUG: 36,
                logging.INFO: 32,
            }.get(record.levelno, 0)
            if settings['debug']:
                self._style._fmt = (  # noqa
                    '\033[{}m%(levelname)s\033[0m: '
                    '%(filename)s, line %(lineno)d, %(message)s'
                ).format(color)
            else:
                self._style._fmt = (  # noqa
                    '\033[{}m%(levelname)s\033[0m: %(message)s'
                ).format(color)
        return super().format(record)


logger = logging.getLogger()
handler = logging.StreamHandler()
handler.setFormatter(Formatter())
logger.addHandler(handler)


def update_verbosity():
    """Set the logging verbosity according to the settings object."""
    if settings['debug']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import os
import time
from gettext import gettext as _
from gettext import ngettext

from gi.repository import GObject, Gtk, Gio, Gdk, GLib, Pango

from soundconverter.util.fileoperations import unquote_filename, vfs_walk
from soundconverter.util.soundfile import SoundFile
from soundconverter.util.taskqueue import TaskQueue
from soundconverter.util.logger import logger
from soundconverter.gstreamer.discoverer import add_discoverers
from soundconverter.util.error import show_error
from soundconverter.interface.notify import notification
from soundconverter.interface.mainloop import gtk_iteration, idle
from soundconverter.util.formatting import format_time


# Names of columns in the file list
MODEL = [
    GObject.TYPE_STRING,  # visible filename
    GObject.TYPE_PYOBJECT,  # soundfile
    GObject.TYPE_FLOAT,  # progress
    GObject.TYPE_STRING,  # status
    GObject.TYPE_STRING,  # complete filename
]

COLUMNS = ['filename']


class FileList:
    """List of files added by the user."""

    # List of MIME types which we accept for drops.
    drop_mime_types = ['text/uri-list', 'text/plain', 'STRING']

    def __init__(self, window, builder):
        self.window = window
        self.discoverers = None
        self.filelist = set()

        self.model = Gtk.ListStore(*MODEL)
        self.progress_cache = {}

        self.widget = builder.get_object('filelist')
        self.widget.props.fixed_height_mode = True
        self.sortedmodel = Gtk.TreeModelSort(model=self.model)
        self.widget.set_model(self.sortedmodel)
        self.sortedmodel.set_sort_column_id(4, Gtk.SortType.ASCENDING)
        self.widget.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        self.widget.drag_dest_set(
            Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY
        )
        targets = [
            (accepted, 0, i) for i, accepted
            in enumerate(self.drop_mime_types)
        ]
        self.widget.drag_dest_set_target_list(targets)

        self.widget.connect('drag-data-received', self.drag_data_received)

        renderer = Gtk.CellRendererProgress()
        column = Gtk.TreeViewColumn(
            'progress',
            renderer,
            value=2,
            text=3,
        )
        column.props.sizing = Gtk.TreeViewColumnSizing.FIXED
        self.widget.append_column(column)
        self.progress_column = column
        self.progress_column.set_visible(False)

        renderer = Gtk.CellRendererText()
        renderer.set_property('ellipsize', Pango.EllipsizeMode.MIDDLE)
        column = Gtk.TreeViewColumn(
            'Filename',
            renderer,
            markup=0,
        )
        column.props.sizing = Gtk.TreeViewColumnSizing.FIXED
        column.set_expand(True)
        self.widget.append_column(column)

        self.window.progressbarstatus.hide()

        self.invalid_files_list = []
        self.good_uris = []

    def drag_data_received(self, widget, context, x, y, selection, mime_id, time):
        widget.stop_emission('drag-data-received')
        if 0 <= mime_id < len(self.drop_mime_types):
            text = selection.get_data().decode('utf-8')
            uris = [uri.strip() for uri in text.split('\n')]
            self.add_uris(uris)
            context.finish(True, False, time)

    def get_files(self):
        """Return all valid SoundFile objects."""
        return [i[1] for i in self.sortedmodel]

    @idle
    def add_uris(self, uris, base=None, extensions=None):
        """Add URIs that should be converted to the list in the GTK interface.

        uris is a list of string URIs, which are absolute paths
        starting with 'file://'

        extensions is a list of strings like ['.ogg', '.oga'],
        in which case only files of this type are added to the
        list. This can be useful when files of multiple types
        are inside a directory and only some of them should be
        converted. Default:None which accepts all types.
        """
        self.progress_cache = {}

        if len(uris) == 0:
            return

        start_t = time.time()
        files = []
        self.window.set_status(_('Scanning files…'))
        # for whichever reason, that set_status needs some more iterations
        # to show up:
        gtk_iteration(True)
        self.window.progressbarstatus.show()
        self.window.progressbarstatus.set_fraction(0)

        for uri in uris:
            gtk_iteration()
            if not uri:
                continue
            if uri.startswith('cdda:'):
                show_error(
                    'Cannot read from Audio CD.',
                    'Use SoundJuicer Audio CD Extractor instead.'
                )
                return
            info = Gio.file_parse_name(uri).query_file_type(
                Gio.FileMonitorFlags.NONE, None
            )
            if info == Gio.FileType.DIRECTORY:
                logger.info('walking: \'{}\''.format(uri))
                if len(uris) == 1:
                    # if only one folder is passed to the function,
                    # use its parent as base path.
                    base = os.path.dirname(uri)

                # get a list of all the files as URIs in
                # that directory and its subdirectories
                filelist = vfs_walk(uri)

                accepted = []
                if extensions:
                    for filename in filelist:
                        for extension in extensions:
                            if filename.lower().endswith(extension):
                                accepted.append(filename)
                    filelist = accepted
                files.extend(filelist)
            else:
                files.append(uri)

        files = [f for f in files if not f.endswith('~SC~')]

        if len(files) == 0:
            show_error('No files found!', '')

        if not base:
            base = os.path.commonprefix(files)
            if base and not base.endswith('/'):
                # we want a common folder
                base = base[0:base.rfind('/')]
                base += '/'
        else:
            base += '/'

        scan_t = time.time()
        logger.info('analysing file integrity')

        # self.good_uris will be populated
        # by the discoverer.
        # It is a list of uris and only contains those files
        # that can be handled by gstreamer
        self.good_uris = []

        self.discoverers = TaskQueue()
        sound_files = []
        for filename in files:
            sound_file = SoundFile(filename, base)
            sound_files.append(sound_file)

        add_discoverers(self.discoverers, sound_files)

        self.discoverers.connect('done', self.discoverer_queue_ended)
        self.discoverers.run()

        self.window.set_status('{}'.format(_('Adding Files…')))
        logger.info('adding: {} files'.format(len(files)))

        # show progress and enable GTK main loop iterations
        # so that the ui stays responsive
        self.window.progressbarstatus.set_text('0/{}'.format(len(files)))
        self.window.progressbarstatus.set_show_text(True)

        while self.discoverers.running:
            progress = self.discoverers.get_progress()[0]
            if progress:
                completed = int(progress * len(files))
                self.window.progressbarstatus.set_fraction(progress)
                self.window.progressbarstatus.set_text(
                    '{}/{}'.format(completed, len(files))
                )
            gtk_iteration()
        logger.info('Discovered {} audiofiles in {} s'.format(
            len(files), round(self.discoverers.get_duration(), 1)
        ))

        self.window.progressbarstatus.set_show_text(False)

        # see if one of the files with an audio extension
        # was not readable.
        known_audio_types = [
            '.flac', '.mp3', '.aac',
            '.m4a', '.mpeg', '.opus', '.vorbis', '.ogg', '.wav'
        ]

        # invalid_files is the number of files that are not
        # added to the list in the current function call
        invalid_files = 0
        # out of those files, that many have an audio file extension
        broken_audiofiles = 0

        sound_files = []
        for discoverer in self.discoverers.all_tasks:
            sound_files += discoverer.sound_files

        for sound_file in sound_files:
            # create a list of human readable file paths
            # that were not added to the list
            if not sound_file.readable:
                filename = sound_file.filename

                extension = os.path.splitext(filename)[1].lower()
                if extension in known_audio_types:
                    broken_audiofiles += 1

                subfolders = sound_file.subfolders
                relative_path = os.path.join(subfolders, filename)

                self.invalid_files_list.append(relative_path)
                invalid_files += 1
                continue
            if sound_file.uri in self.filelist:
                logger.info('file already present: \'{}\''.format(
                    sound_file.uri
                ))
                continue
            self.append_file(sound_file)

        if invalid_files > 0:
            self.window.invalid_files_button.set_visible(True)
            if len(files) == invalid_files == 1:
                # case 1: the single file that should be added is not supported
                show_error(
                    _('The specified file is not supported!'),
                    _('Either because it is broken or not an audio file.')
                )

            elif len(files) == invalid_files:
                # case 2: all files that should be added cannot be added
                show_error(
                    _('All {} specified files are not supported!').format(
                        len(files)
                    ),
                    _('Either because they are broken or not audio files.')
                )

            else:
                # case 3: some files could not be added (that can already be
                # because there is a single picture in a folder of hundreds
                # of sound files). Show an error if this skipped file has a
                # soundfile extension, otherwise don't bother the user.
                logger.info(
                    '{} of {} files were not added to the list'.format(
                        invalid_files, len(files)
                    )
                )
                if broken_audiofiles > 0:
                    show_error(
                        ngettext(
                            'One audio file could not be read by GStreamer!',
                            '{} audio files could not be read by GStreamer!',
                            broken_audiofiles
                        ).format(broken_audiofiles),
                        _(
                            'Check "Invalid Files" in the menu for more '
                            'information.'
                        )
                    )
        else:
            # case 4: all files were successfully added. No error message
            pass

        self.window.set_status()
        self.window.progressbarstatus.hide()
        end_t = time.time()
        logger.debug(
            'Added %d files in %.2fs (scan %.2fs, add %.2fs)' % (
                len(files), end_t - start_t, scan_t - start_t, end_t - scan_t
            )
        )

    def discoverer_queue_ended(self, queue):
        # all tasks done
        self.window.set_sensitive()
        self.window.conversion_ended()

        total_time = queue.get_duration()
        msg = _('Tasks done in %s') % format_time(total_time)

        errors = [
            task.error for task in queue.done
            if task.error is not None
        ]
        if len(errors) > 0:
            msg += ', {} error(s)'.format(len(errors))

        self.window.set_status(msg)
        if not self.window.is_active():
            notification(msg)

        readable = []
        for discoverer in self.discoverers.all_tasks:
            for sound_file in discoverer.sound_files:
                if sound_file.readable:
                    readable.append(sound_file)

        self.good_uris = [sound_file.uri for sound_file in readable]
        self.window.set_status()
        self.window.progressbarstatus.hide()

    def cancel(self):
        if self.discoverers is not None:
            self.discoverers.cancel()

    def format_cell(self, sound_file):
        """Take a SoundFile and return a human readable path to it."""
        return GLib.markup_escape_text(unquote_filename(sound_file.filename))

    def set_row_progress(self, number, progress):
        """Update the progress bar of a single row/file."""
        # when convertin a lot of files updating all progress bars really becomes
        # quite an expensive task
        cached = self.progress_cache.get(number, 0)
        # - progress_cache is faster than self.model for this optimization
        # - skip small changes
        # - make sure it will be set to 1 even if the change is small
        if (progress == 1 and cached != 1) or abs(cached - progress) > 0.02:
            self.model[number][2] = progress * 100.0
            self.progress_cache[number] = progress
            return

    def hide_row_progress(self):
        self.progress_column.set_visible(False)

    def show_row_progress(self):
        self.progress_column.set_visible(True)

    def append_file(self, sound_file):
        """Add a valid SoundFile object to the list of files in the GUI.

        Parameters
        ----------
        sound_file : SoundFile
            This soundfile is expected to be readable by gstreamer
        """
        self.model.append([
            self.format_cell(sound_file), sound_file, 0.0, '', sound_file.uri
        ])
        self.filelist.add(sound_file.uri)
        sound_file.filelist_row = len(self.model) - 1

    def remove(self, iterator):
        uri = self.model.get(iterator, 1)[0].uri
        self.filelist.remove(uri)
        self.model.remove(iterator)

    def is_nonempty(self):
        try:
            self.model.get_iter((0,))
        except ValueError:
            return False
        return True



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"""Functions for formatting strings."""


def format_time(seconds):
    units = [(86400, 'd'),
             (3600, 'h'),
             (60, 'm'),
             (1, 's')]
    seconds = round(seconds)
    result = []
    for factor, unity in units:
        count = int(seconds / factor)
        seconds -= count * factor
        if count > 0 or (factor == 1 and not result):
            result.append('{} {}'.format(count, unity))
    assert seconds == 0
    return ' '.join(result)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA


class GladeWindow:
    callbacks = {}
    builder = None

    def __init__(self, builder):
        """Init GladeWindow, store the objects's potential callbacks for later.

        You have to call connect_signals() when all descendants are ready.
        """
        GladeWindow.builder = builder
        GladeWindow.callbacks.update(dict(
            [[x, getattr(self, x)] for x in dir(self) if x.startswith('on_')]
        ))

    def __getattr__(self, attribute):
        """Allow direct use of window widget."""
        widget = GladeWindow.builder.get_object(attribute)
        if widget is None:
            raise AttributeError('Widget \'{}\' not found'.format(attribute))
        self.__dict__[attribute] = widget  # cache result
        return widget

    @staticmethod
    def connect_signals():
        """Connect all GladeWindow objects to theirs respective signals."""
        GladeWindow.builder.connect_signals(GladeWindow.callbacks)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"""Utils for generating names."""

import string
import time
import re
import os
from random import random
import urllib.request
import urllib.parse
import urllib.error
import unicodedata
from gettext import gettext as _
from soundconverter.util.fileoperations import vfs_exists, filename_to_uri, \
    split_uri, is_uri, beautify_uri
from soundconverter.util.settings import get_gio_settings
from soundconverter.util.formats import get_file_extension

basename_patterns = [
    ('{inputname}', _('Same as input, but replacing the suffix')),
    ('{inputname}.{ext}', _('Same as input, but with an additional suffix')),
    ('{track-number:02}-{title}', _('Track number - title')),
    ('{title}', _('Track title')),
    ('{artist}-{title}', _('Artist - title')),
    ('Custom', _('Custom filename pattern')),
]

subfolder_patterns = [
    ('{album-artist}/{album}', _('artist/album')),
    ('{album-artist}-{album}', _('artist-album')),
    ('{album-artist} - {album}', _('artist - album')),
]

# custom filename patterns
english_patterns = 'Artist Album Album-Artist Title Track Total Genre Date ' \
                   'Year Timestamp DiscNumber DiscTotal Ext'

# traductors: These are the custom filename patterns. Only if it makes sense.
locale_patterns = _('Artist Album Album-Artist Title Track Total Genre Date '
                    'Year Timestamp DiscNumber DiscTotal Ext')

patterns_formats = (
    '{artist}',
    '{album}',
    '{album-artist}',
    '{title}',
    '{track-number:02}',
    '{track-count:02}',
    '{genre}',
    '{date}',
    '{year}',
    '{timestamp}',
    '{album-disc-number}',
    '{album-disc-count}',
    '{target-ext}',
)

# Name and pattern for CustomFileChooser
filepattern = (
    (_('All files'), '*.*'),
    ('MP3', '*.mp3'),
    ('Ogg Vorbis', '*.ogg;*.oga'),
    ('iTunes AAC ', '*.m4a'),
    ('Windows WAV', '*.wav'),
    ('AAC', '*.aac'),
    ('FLAC', '*.flac'),
    ('AC3', '*.ac3')
)

# Example: { 'artist': '{Artist}' }
locale_patterns_dict = dict(list(zip(
    [p.lower() for p in english_patterns.split()],
    ['{%s}' % p for p in locale_patterns.split()]
)))

# add english and locale
custom_patterns = english_patterns + ' ' + locale_patterns
# convert to list
custom_patterns = ['{%s}' % p for p in custom_patterns.split()]
# and finally to dict, thus removing doubles.
# Example: { '{Künstler}': '{artist}' } depending on the locale
custom_patterns = dict(list(zip(custom_patterns, patterns_formats * 2)))


def process_custom_pattern(pattern):
    """Convert e.g. '{Künster}' to '{artist}' in the pattern."""
    for locale_name, tag_name in custom_patterns.items():
        pattern = pattern.replace(locale_name, tag_name)
    return pattern


def get_basename_pattern():
    """Get the currently selected or custom filename pattern.

    For example '{artist}-{title}', without target extension.

    A custom-filename-pattern can also serve the purpose of a subfolder-pattern
    by having forward slashes.
    """
    settings = get_gio_settings()

    index = settings.get_int('name-pattern-index')
    if index >= len(basename_patterns) or index < -1:
        index = 0

    if index in (-1, len(basename_patterns) - 1):
        pattern = settings.get_string('custom-filename-pattern')
        return process_custom_pattern(pattern)

    # an index of -1 selects the last entry on purpose
    return basename_patterns[index][0]


def get_subfolder_pattern():
    """Get the currently selected subfolder pattern.

    For example '{album-artist}/{album}', to create those new
    subfolders in the slected_folder based on tags.
    """
    # Since it is not a free text input on the ui, process_custom_pattern
    # doesn't have to be used
    settings = get_gio_settings()
    index = settings.get_int('subfolder-pattern-index')
    if index >= len(subfolder_patterns) or index < -1:
        index = 0
    return subfolder_patterns[index][0]


class TargetNameGenerator:
    """Generator for creating the target name from an input name.

    Create this class every time when the queue for conversion starts,
    because it remembers all relevant settings to avoid affecting the name
    generation of a running conversion by changing them in the ui.

    This class, once created, can create the names for all conversions in the
    queue, there is no need to create one TargetNameGenerator per Converter.
    """
    def __init__(self):
        # remember settings from when TargetNameGenerator was created:
        settings = get_gio_settings()
        self.same_folder_as_input = settings.get_boolean('same-folder-as-input')
        self.selected_folder = settings.get_string('selected-folder')
        self.output_mime_type = settings.get_string('output-mime-type')
        self.vorbis_oga_extension = settings.get_boolean('vorbis-oga-extension')
        self.create_subfolders = settings.get_boolean('create-subfolders')
        self.replace_messy_chars = settings.get_boolean('replace-messy-chars')
        self.subfolder_pattern = get_subfolder_pattern()
        self.basename_pattern = get_basename_pattern()
        self.suffix = get_file_extension(self.output_mime_type)

        # Enforcing such rules helps to avoid undefined and untested
        # behaviour of functions and to reduce their complexity.
        if not self.same_folder_as_input and not is_uri(self.selected_folder):
            raise ValueError(
                'your selected folder {} should be in URI format.'.format(
                    self.selected_folder
                )
            )
        # If you wish to provide an uri scheme like ftp:// it should be in
        # the selected_folder instead.
        if is_uri(self.basename_pattern) or is_uri(self.subfolder_pattern):
            raise ValueError(
                'patterns should be patterns, not complete URIs.'
            )

    @staticmethod
    def _unicode_to_ascii(unicode_string):
        # thanks to http://code.activestate.com/recipes/251871/
        return str(unicodedata.normalize('NFKD', unicode_string).encode('ASCII', 'ignore'), 'ASCII')

    @staticmethod
    def safe_string(name):
        """Replace all special characters in a string.

        Replace all characters that are not ascii, digits or '.' '-' '_' '/'
        with '_'. Umlaute will be changed to their closest non-umlaut
        counterpart.

        Will not break URI schemes.
        """
        scheme, name = split_uri(name)
        nice_chars = string.ascii_letters + string.digits + '.-_/'
        return (scheme or '') + ''.join([
            c if c in nice_chars else '_' for c in name
        ])

    @staticmethod
    def safe_uri(parent, child):
        """Replace all unusual characters in child for non-existing folders.

        Replace all characters that are not ascii, digits or '.' '-' '_' '/'
        with '_'. Umlaute will be changed to their closest non-umlaut
        counterpart. Will not be applied on the part of the path that already
        exists, as that part apparently is already safe.

        Returns an URI.

        Parameters
        ----------
        parent : string
            Part of filename starting from the beginning of it that should be
            considered safe already and not further modified. Has to be an
            URI. For example "file:///home/user/"
        child : string
            Part of the path that needs to be modified to be safe.
            May already be URI escaped, but doesn't have to.
            Substrings of this path may already exist, in which case they
            will not be modified (" " won't be replaced with "_" then, They
            will be escaped to %20 though to fit the URI format)
            For example "music/artist/file.mp3"
        """
        # some validation of input parameters
        if not is_uri(parent):
            raise ValueError('expected the parent to be an URI')
        if child.startswith(parent):
            raise ValueError(
                'wrong usage. Child "{}" should be the '.format(child) +
                'child of parent "{}", not the whole path'.format(child)
            )
        if child.startswith('./'):
            raise ValueError(
                'you cannot provide a relative path to the cwd starting '
                'with "./", the child is intended to be relative to the '
                'parent.'
            )
        if len(child) == 0:
            raise ValueError('empty filename')
        if is_uri(child):
            raise ValueError(
                'expected child "{}" to be a subfolder, '.format(child) +
                'path not a complete URI.'
            )

        # ensure URI escaping of the parent
        parent = filename_to_uri(parent)

        # ensure correct slahes like "file:///parent/" "subdir1/subdir2"
        if not parent.endswith('/'):
            parent += '/'
        if child.startswith('/'):
            child = child[1:]

        # figure out how much of the path already exists
        # split into for example [/test', '/baz.flac'] or ['qux.mp3']
        split = [s for s in re.split(r'((?:/|^)[^/]+)', child) if s != '']
        safe = ''
        while len(split) > 0:
            part = split.pop(0)
            if vfs_exists(parent + safe + part):
                safe += part
            else:
                # put the remaining unknown non-existing path back together
                # and make it safe
                rest = part + ''.join(split)

                # the path is in uri format, so before applying safe_string,
                # unquote it.
                # otherwise '%20' becomes '_20' when it should be '_'
                rest = urllib.parse.unquote(rest)
                rest = TargetNameGenerator._unicode_to_ascii(rest)
                rest = TargetNameGenerator.safe_string(rest)
                safe += rest
                break

        if parent:
            safe = parent + safe

        # ensure URI escaping of the result
        safe = filename_to_uri(safe)

        return safe

    def find_format_string_tags(self, pattern):
        """Given a pattern find all tags that can be substituted.

        For example for {i}{a}b{c}{{d}}{e}c{{{g}}}{hhh} it should
        find i, a, c, e, g, hhh
        """
        variables = []
        candidates = re.findall(r'{+.+?}+', pattern)
        for candidate in candidates:
            # check if the candidate is made out of an uneven number of
            # curly braces. Two curly braces are a single escaped one.
            if not re.match(r'^(?:{{)*([^{}]+?)(?:}})*$', candidate):
                # get the inner curly brace content
                variable = re.search(r'([^{}]+)', candidate)[1]
                variables.append(variable)
        return variables

    def fill_pattern(self, sound_file, pattern):
        """Fill tags into a filename pattern for a SoundFile.

        Parameters
        ----------
        sound_file : SoundFile
            The soundfile has to have tags, which can be done with the
            discoverer task.
        pattern : string
            For example '{album-artist}/{album}/{title}'. Should not
            be an URI
        """
        tags = sound_file.tags

        filename = beautify_uri(sound_file.uri)
        filename = os.path.basename(filename)
        filename, ext = os.path.splitext(filename)
        assert '/' not in filename
        mapping = {
            'filename': filename,
            'inputname': filename,
            'ext': ext[1:],
            'target-ext': self.suffix,
            'album': _('Unknown Album'),
            'artist': _('Unknown Artist'),
            'album-artist': _('Unknown Artist'),
            'title': filename,
            'track-number': 0,
            'track-count': 0,
            'genre': _('Unknown Genre'),
            'year': _('Unknown Year'),
            'date': _('Unknown Date'),
            'album-disc-number': 0,
            'album-disc-count': 0,
        }

        for key in tags:
            mapping[key] = tags[key]
            if isinstance(mapping[key], str):
                # take care of tags containing slashes
                mapping[key] = mapping[key].replace('/', '-')
                if key.endswith('-number'):
                    mapping[key] = int(mapping[key])

        # when artist set & album-artist not, use artist for album-artist
        if 'artist' in tags and 'album-artist' not in tags:
            mapping['album-artist'] = tags['artist']

        # add timestamp to substitution dict -- this could be split into more
        # entries for more fine-grained control over the string by the user...
        timestamp_string = time.strftime('%Y%m%d_%H_%M_%S')
        mapping['timestamp'] = timestamp_string

        # if some tag is called "venue", map it to "Unknown Venue"
        tags_in_pattern = self.find_format_string_tags(pattern)
        unknown_tags = [tag for tag in tags_in_pattern if tag not in mapping]
        for tag in unknown_tags:
            mapping[tag] = 'Unknown {}'.format(tag.capitalize())

        # now fill the tags in the pattern with values:
        result = pattern.format(**mapping)

        return result

    def generate_temp_path(self, sound_file):
        """Generate a random filename that doesn't exist yet."""
        basename = os.path.split(sound_file.uri)[1]
        parent_uri = self._get_common_target_uri(sound_file)
        while True:
            rand = str(random())[-6:]
            if self.replace_messy_chars:
                filename = basename + '~' + rand + '~SC~'
                final_uri = TargetNameGenerator.safe_uri(parent_uri, filename)
            else:
                final_uri = parent_uri + basename + '~' + rand + '~SC~'
            if not vfs_exists(final_uri):
                return final_uri

    def _get_target_subfolder(self, sound_file):
        """Get subfolders that should be created for the target file.

        Might be None. They may also already exist, for example because a
        previous conversion created them.
        """
        basename_pattern = self.basename_pattern
        subfolder_pattern = self.subfolder_pattern

        subfolder = None
        if self.create_subfolders:
            subfolder = self.fill_pattern(sound_file, subfolder_pattern)
        elif sound_file.subfolders is not None and '/' not in basename_pattern:
            # use existing subfolders between base_path and the soundfile, but
            # only if the basename_pattern does not create subfolders.
            # For example:
            # .subfolders may have an existing structure of artist/album,
            # whereas basename might create a new structure of year/artist
            subfolder = sound_file.subfolders
        return subfolder

    def _get_common_target_uri(self, sound_file):
        """Get the directory into which all files are converted.

        Ends with '/'
        """
        if self.same_folder_as_input:
            parent = sound_file.base_path
        else:
            parent = self.selected_folder
        if not parent.endswith('/'):
            parent += '/'
        # ensure it is an URI, possibly adding file://
        return filename_to_uri(parent)

    def _get_target_filename(self, sound_file):
        """Get the output filename for the soundfile."""
        # note, that basename_pattern might actually contain subfolders, so
        # it's not always only a basename.
        # It's the deepest part of the target path though.
        return '{}.{}'.format(
            self.fill_pattern(sound_file, self.basename_pattern),
            self.suffix
        )

    def generate_target_uri(self, sound_file, for_display=False):
        """Generate a target filename in URI format based on the settings.

        Patterns will be populated with tags.

        Parameters
        ----------
        sound_file : SoundFile
        for_display : bool
            Format it nicely in order to print it somewhere
        """
        # the beginning of the uri that all soundfiles will have in common
        # does not need to be processed in safe_uri.
        parent_uri = self._get_common_target_uri(sound_file)
        # custom subfolders and such, changes depending on the soundfile
        subfolder = self._get_target_subfolder(sound_file)
        # filename, also changes depending on the soundfile
        filename = self._get_target_filename(sound_file)

        # put together
        if subfolder is not None:
            child = os.path.join(subfolder, filename)
        else:
            child = filename
        # child should be quoted to form a proper URI together with parent
        child = urllib.parse.quote(child)

        if child.startswith('/'):
            # os.path.join will omit parent_uri otherwise
            child = child[1:]

        # subfolder and basename need to be cleaned
        if self.replace_messy_chars:
            path = self.safe_uri(parent_uri, child)
        else:
            path = os.path.join(parent_uri, child)

        if for_display:
            return beautify_uri(path)
        else:
            return path



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA


from soundconverter.util.settings import get_gio_settings
from soundconverter.util.logger import logger


filename_denylist = (
    '*.iso',
)


def get_mime_type_mapping():
    """Return a mapping of file extension to mime type."""
    mime_types = {
        'ogg': 'audio/x-vorbis', 'flac': 'audio/x-flac', 'wav': 'audio/x-wav',
        'mp3': 'audio/mpeg', 'aac': 'audio/x-m4a', 'm4a': 'audio/x-m4a',
        'opus': 'audio/ogg; codecs=opus'
    }
    return mime_types


def get_mime_type(audio_format):
    """Return the matching mime-type or None if it is not supported.

    Parameters
    ----------
    audio_format : string
        for example 'mp3' or 'audio/mpeg' for which the result would
        be 'audio/mpeg'
    """
    mime_types = get_mime_type_mapping()
    if audio_format not in mime_types.values():
        # possibly a file extension
        return mime_types.get(audio_format, None)
    else:
        # already a mime string
        return audio_format


def get_file_extension(mime):
    """Return the matching file extension or '?' if it is not supported.

    Examples: 'mp3', 'flac'.

    Parameters
    ----------
    mime : string
        mime string (like 'audio/x-m4a')
    """
    mime_types = get_mime_type_mapping()
    if mime in mime_types:
        # already an extension
        suffix = mime
        if suffix == 'ogg':
            if get_gio_settings().get_boolean('vorbis-oga-extension'):
                suffix = 'oga'
        return suffix
    else:
        mime2ext = {mime: ext for ext, mime in mime_types.items()}
        return mime2ext.get(mime, '?')


def get_quality_setting_name():
    """Get the settings name for quality for the set output-mime-type."""
    settings = get_gio_settings()
    mime_type = settings.get_string('output-mime-type')
    if mime_type == 'audio/mpeg':
        mode = settings.get_string('mp3-mode')
        setting_name = {
            'cbr': 'mp3-cbr-quality',
            'abr': 'mp3-abr-quality',
            'vbr': 'mp3-vbr-quality'
        }[mode]
    else:
        setting_name = {
            'audio/x-vorbis': 'vorbis-quality',
            'audio/x-m4a': 'aac-quality',
            'audio/ogg; codecs=opus': 'opus-bitrate',
            'audio/x-flac': 'flac-compression',
            'audio/x-wav': 'wav-sample-width'
        }[mime_type]
    return setting_name


def get_bitrate_from_settings():
    """Get a human readable bitrate from quality settings.

    For example '~224 kbps'
    """
    settings = get_gio_settings()
    mime_type = settings.get_string('output-mime-type')
    mode = settings.get_string('mp3-mode')

    bitrate = 0
    approx = True

    if mime_type == 'audio/x-vorbis':
        quality = max(0, min(1, settings.get_double('vorbis-quality'))) * 10
        quality = round(quality)
        bitrates = (64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 500)
        bitrate = bitrates[quality]

    elif mime_type == 'audio/x-m4a':
        bitrate = settings.get_int('aac-quality')

    elif mime_type == 'audio/ogg; codecs=opus':
        bitrate = settings.get_int('opus-bitrate')

    elif mime_type == 'audio/mpeg':
        mp3_quality_setting_name = {
            'cbr': 'mp3-cbr-quality',
            'abr': 'mp3-abr-quality',
            'vbr': 'mp3-vbr-quality'
        }[mode]
        setting = settings.get_int(mp3_quality_setting_name)
        if mode == 'vbr':
            # depends on the input audio
            return 'N/A'
        if mode == 'cbr':
            approx = False
            bitrate = setting
        if mode == 'abr':
            bitrate = setting

    elif mime_type == 'audio/x-wav':
        approx = False
        output_resample = settings.get_boolean('output-resample')
        resample_rate = settings.get_int('resample-rate')
        sample_width = settings.get_int('wav-sample-width')
        if output_resample:
            bitrate = sample_width * resample_rate / 1000
        else:
            # the actual bitrate will depend on the input audio, which
            # cannot be known in the settings menu beforehand. Assume 44100
            # which is the most common.
            bitrate = sample_width * 44100 / 1000

    if bitrate:
        if approx:
            return '~{} kbps'.format(bitrate)
        else:
            return '{} kbps'.format(bitrate)
    else:
        return 'N/A'


def get_default_quality(mime, mode='vbr'):
    """Return a default quality if the -q parameter is not set.

    Parameters
    ----------
    mime : string
        mime type
    mode : string
        one of 'cbr', 'abr' and 'vbr' for mp3
    """
    # get 6-tuple of qualities
    default = {
        'audio/x-vorbis': 1.0,
        'audio/x-m4a': 400,
        'audio/ogg; codecs=opus': 192,
        'audio/mpeg': {
            'cbr': 320,
            'abr': 320,
            'vbr': 0,  # inverted !
        },
        'audio/x-wav': 16,
        'audio/x-flac': 5
    }[mime]

    if isinstance(default, dict):
        default = default[mode]

    return default


def get_quality(mime, value, mode='vbr', reverse=False):
    """Map an integer between 0 and 5 to a proper quality/compression value.

    Parameters
    ----------
    mime : string
        mime type
    value : number
        between 0 and 5, or 0 and 2 for flac and wav. -1 indexes the highest
        quality.
    mode : string
        one of 'cbr', 'abr' and 'vbr' for mp3
    reverse : bool
        default False. If True, this function returns the original
        value-parameter given a quality setting. Value becomes the input for
        the quality then.
    """

    # get 6-tuple of qualities
    qualities = {
        'audio/x-vorbis': (0.0, 0.2, 0.4, 0.6, 0.8, 1.0),
        'audio/x-m4a': (64, 96, 128, 192, 256, 320),
        'audio/ogg; codecs=opus': (48, 64, 96, 128, 160, 192),
        'audio/mpeg': {
            'cbr': (64, 96, 128, 192, 256, 320),
            'abr': (64, 96, 128, 192, 256, 320),
            'vbr': (9, 7, 5, 3, 1, 0),  # inverted !
        },
        'audio/x-wav': (8, 16, 32),
        'audio/x-flac': (0, 5, 8)
    }[mime]

    if isinstance(qualities, dict):
        qualities = qualities[mode]

    # return depending on function parameters
    if reverse:
        if isinstance(value, float):
            # floats are inaccurate, search for close value
            for i, quality in enumerate(qualities):
                if abs(value - quality) < 0.01:
                    return i

        if value in qualities:
            return qualities.index(value)

        # might be some custom value set e.g. in batch mode.
        # the reverse mode is only interesting for the ui though, because
        # it has predefined qualities as opposed to batch. So this is
        # either a setting leaking from some tests or the batch mode
        # persisted something.
        if mime == 'mp3':
            ftype_mode = '{} {}'.format(mime, mode)
        else:
            ftype_mode = mime
        logger.warning(
            'tried to index unknow {} quality {}'.format(ftype_mode, value)
        )
        return None
    else:
        # normal index
        if value > len(qualities):
            raise ValueError('quality index {} has to be < {}'.format(
                value, len(qualities)
            ))
        return qualities[value]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA


"""Holds all the settings for both cli arguments as well as Gio settings.

Gio Settings are set in the UI and retained over restarts. In batch mode,
some of the CLI args are written into that temporarily.
"""

from multiprocessing import cpu_count
from gi.repository import Gio


# Use get_gio_settings instead of importing this directly, because this
# object changes in tests, and also this object will be replaced with a
# memory backend for the batch mode.
_gio_settings = Gio.Settings(schema='org.soundconverter')  # do not import!


def get_gio_settings():
    """Return the current Gio.Settings object.

    Use this isntead of importing _gio_settings directly.
    """
    return _gio_settings


# Arguments that can exclusively set over the CLI
settings = {
    'main': 'gui',
    'debug': False
}


def set_gio_settings(settings):
    """Overwrite the default Gio.Settings object.

    For example use a memory backend instead.

    Parameters
    ----------
    settings : Gio.Settings
        You can get this by using for example Gio.new_with_backend or
        Gio.Settings
    """
    global _gio_settings
    _gio_settings = settings


def get_num_jobs():
    """Return the number of jobs that should be run in parallel."""
    return (
        _gio_settings.get_int('number-of-jobs')
        if _gio_settings.get_boolean('limit-jobs')
        else cpu_count()
    )



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA


from gi.repository import GObject


class Task(GObject.Object):
    """Abstract class of a single task."""
    def __init__(self):
        super().__init__()

    # avoid storing a variable called timer in your inheriting class
    def get_progress(self):
        """Fraction of how much of the task is completed.

        Returns a tuple of (progress, weight), because some tasks may
        take longer than others (because it processes more audio), which
        cannot be reflected by the progress alone. The weight might
        correspond to the length of the audio files for example.
        """
        raise NotImplementedError()

    def cancel(self):
        """Cancel the execution of the task."""
        raise NotImplementedError()

    def pause(self):
        """Pause the execution of the task."""
        raise NotImplementedError()

    def resume(self):
        """Resume the execution of the task after pausing."""
        raise NotImplementedError()

    def run(self):
        """Run the task."""
        raise NotImplementedError()

    def done(self):
        """Emit a "done" event."""
        self.emit('done')


GObject.signal_new(
    'done',
    Task,
    GObject.SignalFlags.RUN_FIRST,
    None,
    []
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import os
import re
import urllib.request
import urllib.parse
import urllib.error
from gi.repository import Gio

from soundconverter.util.logger import logger


def unquote_filename(filename):
    """Transform an URL encoded filename to a non-encoded one.

    E.g. '%20' will be changed to ' '
    """
    return urllib.parse.unquote(str(filename))


def beautify_uri(uri):
    """Convert an URI to a normal path.

    Also returns the prefix, for example 'file://'"""
    match = split_uri(uri)
    if match[0] is not None:
        # take the path part from the uri
        path = match[1]
        path = unquote_filename(path)
    else:
        # no uri, take as is, return any existing %20 strings without
        # modifying them.
        path = uri
    return path


def vfs_walk(uri):
    """Similar to os.path.walk, but with Gio.

    uri -- the base folder uri.
    return a list of uri.
    """
    filelist = []

    try:
        dirlist = Gio.file_parse_name(uri).enumerate_children(
            '*', Gio.FileMonitorFlags.NONE, None
        )

        for file_info in dirlist:
            info = dirlist.get_child(file_info).query_file_type(
                Gio.FileMonitorFlags.NONE, None
            )

            uri = dirlist.get_child(file_info).get_uri();

            if info == Gio.FileType.DIRECTORY:
                filelist.extend(vfs_walk(uri))

            if info == Gio.FileType.REGULAR:
                filelist.append(str(uri))
    except Exception as e:
        # this is impossible to write unittests for, because this only happens
        # when the owner of this directory is e.g. root
        logger.error('Failed to walk "%s": "%s"', uri, e)

    return filelist


def vfs_getparent(path):
    """Get folder name."""
    gfile = Gio.file_parse_name(path)
    return gfile.get_parent()


def vfs_unlink(filename):
    """Delete a gnomevfs file."""
    gfile = Gio.file_parse_name(filename)
    return gfile.delete(None)


def vfs_rename(original, newname):
    """Rename a gnomevfs file."""
    gforiginal = Gio.file_parse_name(original)
    gfnew = Gio.file_parse_name(newname)
    if not gfnew.get_parent().query_exists(None):
        fgnew_uri = gfnew.get_parent().get_uri()
        logger.debug('Creating folder: \'{}\''.format(fgnew_uri))
        Gio.File.make_directory_with_parents(gfnew.get_parent(), None)
    gforiginal.move(gfnew, Gio.FileCopyFlags.NONE, None, None, None)


def vfs_exists(filename):
    """Check if file or URI exists."""
    if not is_uri(filename):
        # gio does not support relative path syntax
        filename = os.path.realpath(filename)
    gfile = Gio.file_parse_name(filename)
    return gfile.query_exists(None)


def split_uri(uri):
    """Match a regex to the uri that results in:

    [0]: scheme and authority, might be None if not an uri
    [1]: filename. This still has to be unquoted!
    """
    if not isinstance(uri, str):
        raise ValueError('cannot split {} {}'.format(type(uri), uri))

    match = re.match(r'^([a-zA-Z]+://([^/]+?)?)?(/.*)', uri)
    if match is None:
        # not an uri
        return None, uri
    return match[1], match[3]


def is_uri(uri):
    return split_uri(uri)[0] is not None


def filename_to_uri(filename, prefix='file://'):
    """Convert a filename to a valid uri.

    Parameters
    ----------
    filename : string
        Filename can be a relative or absolute path, or an URI. If an URI,
        only characters that are not escaped yet will be escaped.
    prefix : string
        for example 'file://'
    """
    match = split_uri(filename)
    if match[0]:
        # it's an URI! It can be basically just returned as is. But to make
        # sure that all characters are URI escaped, the path will be
        # escaped again. Don't quote the schema.
        # e.g. a pattern contained file:// in front but inserting tags into it
        # resulted in whitespaces.
        # ' %20' to '  ' to '%20%20'. Don't quote it to '%20%2520'!
        filename = unquote_filename(match[1])
        filename = urllib.parse.quote(filename)
        uri = match[0] + filename
    else:
        # convert to absolute path
        filename = os.path.realpath(filename)
        # it's a normal path. If it happens to contain %25, it might be
        # part of the album name or something. ' %20' should become '%20%2520'
        uri = prefix + urllib.parse.quote(filename)
    return uri


# GStreamer gnomevfssrc helpers


def vfs_encode_filename(filename):
    return filename_to_uri(filename)


def file_encode_filename(filename):
    return Gio.get_local_path_from_uri(filename).replace(' ', r'\ ')



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

from fnmatch import fnmatch
from threading import Thread

from gi.repository import Gst, GObject, GstPbutils, GLib

from soundconverter.util.task import Task
from soundconverter.util.logger import logger
from soundconverter.util.settings import get_num_jobs
from soundconverter.util.formats import filename_denylist

type_getters = {
    GObject.TYPE_STRING: 'get_string',
    GObject.TYPE_DOUBLE: 'get_double',
    GObject.TYPE_FLOAT: 'get_float',
    GObject.TYPE_INT: 'get_int',
    GObject.TYPE_UINT: 'get_uint',
}


def add_discoverers(task_queue, sound_files):
    """Fill a TaskQueue with Discoverer tasks for optimized discovery."""
    chunk = []
    chunksize = len(sound_files) / get_num_jobs()
    for sound_file in sound_files:
        chunk.append(sound_file)
        if len(chunk) >= chunksize or sound_file is sound_files[-1]:
            discoverer = Discoverer(chunk)
            task_queue.add(discoverer)
            chunk = []

    if len(chunk) > 0:
        raise AssertionError(
            'All chunks should have been added to discoverers'
        )


def get_sound_files(task_queue):
    """Get all SoundFiles of discoverer tasks in a TaskQueue"""
    sound_files = []
    for task in task_queue:
        if isinstance(task, Discoverer):
            for sound_file in task.sound_files:
                sound_files.append(sound_file)
    return sound_files


def is_denylisted(sound_file):
    """Check the file against the denylist."""
    for file_pattern in filename_denylist:
        if fnmatch(sound_file.uri, file_pattern):
            return file_pattern
    return False


class DiscovererThread(Thread):
    """Discover if multiple SoundFiles can be read and their tags."""

    # This is the fastest way I could figure out. discover_uri_async
    # was not faster than discover_uri, and the UI only stayed responsive
    # as long as only 1 discover_uri_async job was running at a time.
    # Maybe it was a bit too much for the GLib event loop, who knows.
    # Running multiple discover_uri_async jobs at a time also didn't
    # improve the performance. By using threads with synchronous
    # discovery, I could get a very responsive UI while spawning 12
    # Threads with a drastic ~5-times performance increase for ~360
    # files.
    # I couldn't get it to work with the multiprocessing module though,
    # because the discover_uri function would hang.

    def __init__(self, sound_files, bus):
        super().__init__()
        self.sound_files = sound_files
        self.bus = bus

    def run(self):
        """Run the Thread."""
        for sound_file in self.sound_files:
            self._analyse_file(sound_file)

            msg_type = Gst.MessageType(Gst.MessageType.PROGRESS)
            msg = Gst.Message.new_custom(msg_type, None, None)
            self.bus.post(msg)

        msg_type = Gst.MessageType(Gst.MessageType.EOS)
        msg = Gst.Message.new_custom(msg_type, None, None)
        self.bus.post(msg)

    def _analyse_file(self, sound_file):
        """Figure out readable, tags and duration properties."""
        sound_file.readable = False

        denylisted_pattern = is_denylisted(sound_file)
        if denylisted_pattern:
            logger.info('filename denylisted ({}): {}'.format(
                denylisted_pattern, sound_file.filename_for_display
            ))
            return

        try:
            discoverer = GstPbutils.Discoverer()
            info = discoverer.discover_uri(sound_file.uri)

            # whatever anybody might ever need from it, here it is:
            sound_file.info = info

            # Read root tags
            taglist = info.get_tags()
            if taglist: 
                taglist.foreach(lambda *args: self._add_tag(*args, sound_file))

            for audio_stream in info.get_audio_streams():
                # Read tags for each audio stream
                taglist = audio_stream.get_tags()
                taglist.foreach(lambda *args: self._add_tag(*args, sound_file))

            filename = sound_file.filename_for_display
            logger.debug('found tag: {}'.format(filename))
            for tag, value in sound_file.tags.items():
                logger.debug('    {}: {}'.format(tag, value))

            duration = info.get_duration() / Gst.SECOND
            if duration == 0:
                # might be an image
                return

            # since threads share memory, this doesn't have to be sent
            # over a bus or queue, but rather can be written into the
            # sound_file
            sound_file.duration = duration

            sound_file.readable = True
        except Exception as error:
            if not isinstance(error, GLib.Error):
                logger.error(str(error))

    def _add_tag(self, taglist, tag, sound_file):
        """Convert the taglist to a dict one by one."""
        # only really needed to construct output paths
        tag_type = Gst.tag_get_type(tag)

        if tag_type in type_getters:
            getter = getattr(taglist, type_getters[tag_type])
            value = str(getter(tag)[1])
            sound_file.tags[tag] = value

        if 'datetime' in tag:
            date_time = taglist.get_date_time(tag)[1]
            sound_file.tags['year'] = date_time.get_year()
            sound_file.tags['date'] = date_time.to_iso8601_string()[:10]


class Discoverer(Task):
    """Find type and tags of a SoundFile if possible."""

    def __init__(self, sound_files):
        """Find type and tags of a SoundFile if possible."""
        self.sound_files = sound_files
        self.error = None
        self.running = False
        self.callback = lambda: None
        self.discovered = 0
        self.queue = None

        self.bus = None
        self.thread = None

        super().__init__()

    def get_progress(self):
        """Fraction of how much of the task is completed."""
        return self.discovered / len(self.sound_files), 1

    def cancel(self):
        """Cancel execution of the task."""
        # fast task, use case doesn't exist
        self.done()

    def pause(self):
        """Pause execution of the task."""
        # fast task, use case doesn't exist

    def resume(self):
        """Resume execution of the task."""
        # fast task, use case doesn't exist

    def run(self):
        self.running = True
        bus = Gst.Bus()
        bus.connect('message', self._on_message)
        bus.add_signal_watch()
        thread = DiscovererThread(self.sound_files, bus)
        thread.start()
        self.bus = bus
        self.thread = thread

    def _on_message(self, _, message):
        """Write down that it is finished and call the callback."""
        if message.type == Gst.MessageType.EOS:
            self.running = False
            self.done()
        else:
            self.discovered += 1



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import urllib.request
import urllib.parse
import urllib.error
from gettext import gettext as _

from gi.repository import Gtk, GLib

from soundconverter.util.fileoperations import filename_to_uri, \
    beautify_uri
from soundconverter.util.soundfile import SoundFile
from soundconverter.util.settings import get_gio_settings
from soundconverter.util.formats import get_quality, \
    get_bitrate_from_settings
from soundconverter.util.namegenerator import TargetNameGenerator, \
    subfolder_patterns, basename_patterns, locale_patterns_dict
from soundconverter.util.logger import logger
from soundconverter.gstreamer.converter import available_elements
from soundconverter.interface.gladewindow import GladeWindow


encoders = [
    ('audio/x-vorbis', 'vorbisenc', 'Ogg Vorbis (.ogg)'),
    ('audio/mpeg', 'lamemp3enc', 'MP3 (.mp3)'),
    ('audio/x-flac', 'flacenc', 'FLAC Lossless (.flac)'),
    ('audio/x-wav', 'wavenc', 'MS Wave (.wav)'),
    ('audio/x-m4a', 'fdkaacenc,faac,avenc_aac', 'AAC (.m4a)'),
    ('audio/ogg; codecs=opus', 'opusenc', 'Opus (.opus)'),
]

rates = [8000, 11025, 16000, 22050, 32000, 44100, 48000, 96000, 128000]


class PreferencesDialog(GladeWindow):
    sensitive_names = [
        'vorbis_quality', 'choose_folder', 'create_subfolders',
        'subfolder_pattern', 'jobs_spinbutton', 'resample_hbox',
        'force_mono'
    ]

    def __init__(self, builder, parent):
        self.settings = get_gio_settings()
        GladeWindow.__init__(self, builder)

        self.present_mime_types = []
        self.populate_output_formats()

        self.dialog = builder.get_object('prefsdialog')
        self.dialog.set_transient_for(parent)
        self.example = builder.get_object('example_filename')
        self.force_mono = builder.get_object('force_mono')

        self.target_bitrate = None

        self.sensitive_widgets = {}
        for name in self.sensitive_names:
            self.sensitive_widgets[name] = builder.get_object(name)
            assert self.sensitive_widgets[name] is not None

        self.set_widget_initial_values()
        self.set_sensitive()

        tip = [_('Available patterns:')]
        for k in sorted(locale_patterns_dict.values()):
            tip.append(k)
        self.custom_filename.set_tooltip_text('\n'.join(tip))

        self.output_mime_type.set_id_column(0)

    def populate_output_formats(self):
        """Add the available encoders to the liststore for formats."""
        for mime, encoder_name, display_name in encoders:
            # valid default output?
            encoder_present = any(
                e in available_elements
                for e in encoder_name.split(',')
            )
            if not encoder_present:
                logger.error(
                    '{} {} is not supported, a gstreamer plugins package '
                    'is possibly missing.'.format(mime, encoder_name)
                )
                continue

            # add to supported outputs
            self.present_mime_types.append(mime)
            self.liststore8.append((display_name,))

    def set_widget_initial_values(self):
        self.quality_tabs.set_show_tabs(False)

        if self.settings.get_boolean('same-folder-as-input'):
            widget = self.same_folder_as_input
        else:
            widget = self.into_selected_folder
        widget.set_active(True)

        self.target_folder_chooser = Gtk.FileChooserDialog(
            title=_('Add Folder…'),
            transient_for=self.dialog,
            action=Gtk.FileChooserAction.SELECT_FOLDER
        )

        self.target_folder_chooser.add_button(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL
        )
        self.target_folder_chooser.add_button(
            Gtk.STOCK_OPEN, Gtk.ResponseType.OK
        )

        self.target_folder_chooser.set_select_multiple(False)
        self.target_folder_chooser.set_local_only(False)

        uri = filename_to_uri(urllib.parse.quote(
            self.settings.get_string('selected-folder'), safe='/:@'
        ))
        self.target_folder_chooser.set_uri(uri)
        self.update_selected_folder()

        widget = self.create_subfolders
        widget.set_active(self.settings.get_boolean('create-subfolders'))

        widget = self.subfolder_pattern
        active = self.settings.get_int('subfolder-pattern-index')
        model = widget.get_model()
        model.clear()
        for pattern, desc in subfolder_patterns:
            i = model.append()
            model.set(i, 0, desc)
        widget.set_active(active)

        if self.settings.get_boolean('replace-messy-chars'):
            widget = self.replace_messy_chars
            widget.set_active(True)

        if self.settings.get_boolean('delete-original'):
            self.delete_original.set_active(True)

        current_mime_type = self.settings.get_string('output-mime-type')
        for i, mime in enumerate(self.present_mime_types):
            if current_mime_type == mime:
                self.output_mime_type.set_active(i)
        self.change_mime_type(current_mime_type)

        # display information about mp3 encoding
        if 'lamemp3enc' not in available_elements:
            widget = self.lame_absent
            widget.show()

        widget = self.vorbis_quality
        quality = self.settings.get_double('vorbis-quality')
        quality_setting = get_quality('audio/x-vorbis', quality, reverse=True)
        widget.set_active(-1)
        self.vorbis_quality.set_active(quality_setting)
        if self.settings.get_boolean('vorbis-oga-extension'):
            self.vorbis_oga_extension.set_active(True)

        widget = self.aac_quality
        quality = self.settings.get_int('aac-quality')
        quality_setting = get_quality('audio/x-m4a', quality, reverse=True)
        widget.set_active(quality_setting)

        widget = self.opus_quality
        quality = self.settings.get_int('opus-bitrate')
        quality_setting = get_quality('audio/ogg; codecs=opus', quality, reverse=True)
        widget.set_active(quality_setting)

        widget = self.flac_compression
        quality = self.settings.get_int('flac-compression')
        quality_setting = get_quality('audio/x-flac', quality, reverse=True)
        widget.set_active(quality_setting)

        widget = self.wav_sample_width
        quality = self.settings.get_int('wav-sample-width')
        quality_setting = get_quality('audio/x-wav', quality, reverse=True)
        widget.set_active(quality_setting)

        mode = self.settings.get_string('mp3-mode')
        self.change_mp3_mode(mode)

        widget = self.basename_pattern
        active = self.settings.get_int('name-pattern-index')
        model = widget.get_model()
        model.clear()
        for pattern, desc in basename_patterns:
            iter = model.append()
            model.set(iter, 0, desc)
        widget.set_active(active)

        self.custom_filename.set_text(
            self.settings.get_string('custom-filename-pattern')
        )
        if self.basename_pattern.get_active() == len(basename_patterns)-1:
            self.custom_filename_box.set_sensitive(True)
        else:
            self.custom_filename_box.set_sensitive(False)

        output_resample = self.settings.get_boolean('output-resample')
        self.resample_toggle.set_active(output_resample)

        cell = Gtk.CellRendererText()
        self.resample_rate.pack_start(cell, True)
        self.resample_rate.add_attribute(cell, 'text', 0)
        rate = self.settings.get_int('resample-rate')
        try:
            idx = rates.index(rate)
        except ValueError:
            idx = -1
        self.resample_rate.set_active(idx)

        self.force_mono.set_active(self.settings.get_boolean('force-mono'))

        self.jobs.set_active(self.settings.get_boolean('limit-jobs'))
        self.jobs_spinbutton.set_value(self.settings.get_int('number-of-jobs'))

        self.update_example()

    def update_selected_folder(self):
        self.into_selected_folder.set_use_underline(False)
        self.into_selected_folder.set_label(
            _('Into folder %s') %
            beautify_uri(self.settings.get_string('selected-folder'))
        )

    def update_example(self):
        """Refresh the example in the settings dialog."""
        sound_file = SoundFile('file:///foo/bar.flac')
        sound_file.tags.update({
            'track-number': 1,
            'track-count': 99,
            'album-disc-number': 2,
            'album-disc-count': 9
        })
        sound_file.tags.update(locale_patterns_dict)

        try:
            generator = TargetNameGenerator()
        except ValueError:
            # since this is just for displaying the example we don't
            # care about any errors
            return

        generator.replace_messy_chars = False

        example_path = GLib.markup_escape_text(
            generator.generate_target_uri(sound_file, for_display=True)
        )
        position = 0
        replaces = []

        while True:
            beginning = example_path.find('{', position)
            if beginning == -1:
                break
            end = example_path.find('}', beginning)

            tag = example_path[beginning:end+1]
            available_tags = [
                v.lower() for v in list(locale_patterns_dict.values())
            ]
            if tag.lower() in available_tags:
                bold_tag = tag.replace(
                    '{', '<b>{'
                ).replace(
                    '}', '}</b>'
                )
                replaces.append([tag, bold_tag])
            else:
                red_tag = tag.replace(
                    '{', '<span foreground=\'red\'><i>{'
                ).replace(
                    '}', '}</i></span>'
                )
                replaces.append([tag, red_tag])
            position = beginning + 1

        for tag, formatted in replaces:
            example_path = example_path.replace(tag, formatted)

        self.example.set_markup(example_path)

        markup = '<small>{}</small>'.format(
            _('Target bitrate: %s') % get_bitrate_from_settings()
        )
        self.approx_bitrate.set_markup(markup)

    def set_sensitive(self):
        for widget in list(self.sensitive_widgets.values()):
            widget.set_sensitive(False)

        same_folder = self.settings.get_boolean('same-folder-as-input')
        for name in ['choose_folder', 'create_subfolders',
                     'subfolder_pattern']:
            self.sensitive_widgets[name].set_sensitive(not same_folder)

        self.sensitive_widgets['vorbis_quality'].set_sensitive(
            self.settings.get_string('output-mime-type') == 'audio/x-vorbis')

        self.sensitive_widgets['jobs_spinbutton'].set_sensitive(
            self.settings.get_boolean('limit-jobs'))

        self.sensitive_widgets['resample_hbox'].set_sensitive(True)
        self.sensitive_widgets['force_mono'].set_sensitive(True)

    def run(self):
        self.dialog.run()
        self.dialog.hide()

    def on_delete_original_toggled(self, button):
        self.settings.set_boolean('delete-original', button.get_active())

    def on_same_folder_as_input_toggled(self, button):
        self.settings.set_boolean('same-folder-as-input', True)
        self.set_sensitive()
        self.update_example()

    def on_into_selected_folder_toggled(self, button):
        self.settings.set_boolean('same-folder-as-input', False)
        self.set_sensitive()
        self.update_example()

    def on_choose_folder_clicked(self, button):
        ret = self.target_folder_chooser.run()
        folder = self.target_folder_chooser.get_uri()
        self.target_folder_chooser.hide()
        if ret == Gtk.ResponseType.OK:
            if folder:
                folder = urllib.parse.unquote(folder)
                self.settings.set_string('selected-folder', folder)
                self.update_selected_folder()
                self.update_example()

    def on_create_subfolders_toggled(self, button):
        self.settings.set_boolean('create-subfolders', button.get_active())
        self.update_example()

    def on_subfolder_pattern_changed(self, combobox):
        self.settings.set_int('subfolder-pattern-index', combobox.get_active())
        self.update_example()

    def on_basename_pattern_changed(self, combobox):
        self.settings.set_int('name-pattern-index', combobox.get_active())
        if combobox.get_active() == len(basename_patterns)-1:
            self.custom_filename_box.set_sensitive(True)
        else:
            self.custom_filename_box.set_sensitive(False)
        self.update_example()

    def on_custom_filename_changed(self, entry):
        self.settings.set_string('custom-filename-pattern', entry.get_text())
        self.update_example()

    def on_replace_messy_chars_toggled(self, button):
        self.settings.set_boolean('replace-messy-chars', button.get_active())

    def change_mime_type(self, mime_type):
        """Show the correct quality tab based on the selected format."""
        self.settings.set_string('output-mime-type', mime_type)
        self.set_sensitive()
        self.update_example()
        tabs = {
            'audio/x-vorbis': 0,
            'audio/mpeg': 1,
            'audio/x-flac': 2,
            'audio/x-wav': 3,
            'audio/x-m4a': 4,
            'audio/ogg; codecs=opus': 5,
        }
        self.quality_tabs.set_current_page(tabs[mime_type])

    def on_output_mime_type_changed(self, combo):
        """Called when the format is changed on the UI."""
        selected_display_name = self.liststore8[combo.get_active()][0]
        for mime, _, display_name in encoders:
            if display_name == selected_display_name:
                self.change_mime_type(mime)
                return

    def on_output_mime_type_ogg_vorbis_toggled(self, button):
        if button.get_active():
            self.change_mime_type('audio/x-vorbis')

    def on_output_mime_type_flac_toggled(self, button):
        if button.get_active():
            self.change_mime_type('audio/x-flac')

    def on_output_mime_type_wav_toggled(self, button):
        if button.get_active():
            self.change_mime_type('audio/x-wav')

    def on_output_mime_type_mp3_toggled(self, button):
        if button.get_active():
            self.change_mime_type('audio/mpeg')

    def on_output_mime_type_aac_toggled(self, button):
        if button.get_active():
            self.change_mime_type('audio/x-m4a')

    def on_output_mime_type_opus_toggled(self, button):
        if button.get_active():
            self.change_mime_type('audio/ogg; codecs=opus')

    def on_vorbis_quality_changed(self, combobox):
        if combobox.get_active() == -1:
            return  # just de-selectionning
        fquality = get_quality('audio/x-vorbis', combobox.get_active())
        self.settings.set_double('vorbis-quality', fquality)
        self.update_example()

    def on_vorbis_oga_extension_toggled(self, toggle):
        self.settings.set_boolean('vorbis-oga-extension', toggle.get_active())
        self.update_example()

    def on_aac_quality_changed(self, combobox):
        quality = get_quality('audio/x-m4a', combobox.get_active())
        self.settings.set_int('aac-quality', quality)
        self.update_example()

    def on_opus_quality_changed(self, combobox):
        quality = get_quality('audio/ogg; codecs=opus', combobox.get_active())
        self.settings.set_int('opus-bitrate', quality)
        self.update_example()

    def on_wav_sample_width_changed(self, combobox):
        quality = get_quality('audio/x-wav', combobox.get_active())
        self.settings.set_int('wav-sample-width', quality)
        self.update_example()

    def on_flac_compression_changed(self, combobox):
        quality = get_quality('audio/x-flac', combobox.get_active())
        self.settings.set_int('flac-compression', quality)
        self.update_example()

    def on_force_mono_toggle(self, button):
        self.settings.set_boolean('force-mono', button.get_active())
        self.update_example()

    def change_mp3_mode(self, mode):
        keys = {'cbr': 0, 'abr': 1, 'vbr': 2}
        self.mp3_mode.set_active(keys[mode])

        keys = {
            'cbr': 'mp3-cbr-quality',
            'abr': 'mp3-abr-quality',
            'vbr': 'mp3-vbr-quality',
        }
        quality = self.settings.get_int(keys[mode])

        index = get_quality('audio/mpeg', quality, mode, reverse=True)
        self.mp3_quality.set_active(index)
        self.update_example()

    def on_mp3_mode_changed(self, combobox):
        mode = ('cbr', 'abr', 'vbr')[combobox.get_active()]
        self.settings.set_string('mp3-mode', mode)
        self.change_mp3_mode(mode)

    def on_mp3_quality_changed(self, combobox):
        keys = {
            'cbr': 'mp3-cbr-quality',
            'abr': 'mp3-abr-quality',
            'vbr': 'mp3-vbr-quality'
        }
        mode = self.settings.get_string('mp3-mode')

        bitrate = get_quality('audio/mpeg', combobox.get_active(), mode)
        self.settings.set_int(keys[mode], bitrate)
        self.update_example()

    def on_resample_rate_changed(self, combobox):
        selected = combobox.get_active()
        self.settings.set_int('resample-rate', rates[selected])
        self.update_example()

    def on_resample_toggle(self, rstoggle):
        self.settings.set_boolean('output-resample', rstoggle.get_active())
        self.resample_rate.set_sensitive(rstoggle.get_active())
        self.update_example()

    def on_jobs_toggled(self, jtoggle):
        self.settings.set_boolean('limit-jobs', jtoggle.get_active())
        self.jobs_spinbutton.set_sensitive(jtoggle.get_active())

    def on_jobs_spinbutton_value_changed(self, jspinbutton):
        self.settings.set_int('number-of-jobs', int(jspinbutton.get_value()))




#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import time

from gi.repository import Gtk, GLib


def idle(func):
    def callback(*args, **kwargs):
        GLib.idle_add(func, *args, **kwargs)
    return callback


def gtk_iteration(blocking=False):
    """Keeps the UI and event loops for gst going.

    Paramters
    ---------
    blocking : bool
        If True, will call main_iteration even if no events are pending,
        which will wait until an event is available.
    """
    if blocking:
        while True:
            Gtk.main_iteration()
            if not Gtk.events_pending():
                break
    else:
        while Gtk.events_pending():
            Gtk.main_iteration()


def gtk_sleep(duration):
    """Sleep while keeping the GUI responsive."""
    start = time.time()
    while time.time() < start + duration:
        time.sleep(0.01)
        gtk_iteration()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SoundConverter - GNOME application for converting between audio formats.
# Copyright 2004 Lars Wirzenius
# Copyright 2005-2020 Gautier Portet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import os
from gettext import gettext as _

from gi.repository import Gtk, GLib

from soundconverter.util.fileoperations import filename_to_uri
from soundconverter.util.settings import settings
from soundconverter.util.namegenerator import TargetNameGenerator, \
    filepattern
from soundconverter.util.taskqueue import TaskQueue
from soundconverter.util.logger import logger
from soundconverter.gstreamer.converter import Converter
from soundconverter.util.error import set_error_handler
from soundconverter.util.formatting import format_time
from soundconverter.interface.gladewindow import GladeWindow
from soundconverter.interface.preferences import PreferencesDialog
from soundconverter.interface.filelist import FileList
from soundconverter.interface.mainloop import gtk_iteration, gtk_sleep


class SoundConverterWindow(GladeWindow):
    """Main application class."""

    sensitive_names = [
        'remove', 'clearlist',
        'convert_button'
    ]
    unsensitive_when_converting = [
        'remove', 'clearlist', 'prefs_button',
        'toolbutton_addfile', 'toolbutton_addfolder', 'convert_button',
        'filelist', 'menubar'
    ]

    def __init__(self, builder):
        GladeWindow.__init__(self, builder)

        self.widget = builder.get_object('window')
        self.prefs = PreferencesDialog(builder, self.widget)
        GladeWindow.connect_signals()

        self.filelist = FileList(self, builder)
        self.filelist_selection = self.filelist.widget.get_selection()
        self.filelist_selection.connect('changed', self.selection_changed)
        self.existsdialog = builder.get_object('existsdialog')
        self.existsdialog.message = builder.get_object('exists_message')
        self.existsdialog.apply_to_all = builder.get_object('apply_to_all')

        self.addfolderchooser = Gtk.FileChooserDialog(
            title=_('Add Folder…'),
            transient_for=self.widget,
            action=Gtk.FileChooserAction.SELECT_FOLDER
        )

        self.addfolderchooser.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.addfolderchooser.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)

        self.addfolderchooser.set_select_multiple(True)
        self.addfolderchooser.set_local_only(False)

        self.combo = Gtk.ComboBox()
        self.store = Gtk.ListStore(str)
        self.combo.set_model(self.store)
        combo_rend = Gtk.CellRendererText()
        self.combo.pack_start(combo_rend, True)
        self.combo.add_attribute(combo_rend, 'text', 0)

        for files in filepattern:
            self.store.append(['{} ({})'.format(files[0], files[1])])

        self.combo.set_active(0)
        self.addfolderchooser.set_extra_widget(self.combo)

        self.addchooser = Gtk.FileChooserDialog(
            title=_('Add Files…'),
            transient_for=self.widget,
            action=Gtk.FileChooserAction.OPEN
        )

        self.addchooser.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.addchooser.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)

        self.addchooser.set_select_multiple(True)
        self.addchooser.set_local_only(False)

        self.addfile_combo = Gtk.ComboBox()
        self.addfile_store = Gtk.ListStore(str)
        self.addfile_combo.set_model(self.addfile_store)
        combo_rend = Gtk.CellRendererText()
        self.addfile_combo.pack_start(combo_rend, True)
        self.addfile_combo.add_attribute(combo_rend, 'text', 0)
        self.addfile_combo.connect('changed', self.on_addfile_combo_changed)

        self.pattern = []
        for files in filepattern:
            self.pattern.append(files[1])
            self.addfile_store.append(['{} ({})'.format(files[0], files[1])])

        self.addfile_combo.set_active(0)
        self.addchooser.set_extra_widget(self.addfile_combo)

        # self.aboutdialog.set_property('name', NAME)
        # self.aboutdialog.set_property('version', VERSION)
        # self.aboutdialog.set_transient_for(self.widget)

        self.converter_queue = None

        self.sensitive_widgets = {}
        for name in self.sensitive_names:
            self.sensitive_widgets[name] = builder.get_object(name)
        for name in self.unsensitive_when_converting:
            self.sensitive_widgets[name] = builder.get_object(name)

        self.set_sensitive()
        self.set_status()

    # This bit of code constructs a list of methods for binding to Gtk+
    # signals. This way, we don't have to maintain a list manually,
    # saving editing effort. It's enough to add a method to the suitable
    # class and give the same name in the .glade file.

    def __getattr__(self, attribute):
        """Allow direct use of window widget."""
        widget = self.builder.get_object(attribute)
        if widget is None:
            raise AttributeError('Widget \'{}\' not found'.format(attribute))
        self.__dict__[attribute] = widget  # cache result
        return widget

    def close(self, *args):
        logger.debug('closing…')
        self.filelist.cancel()
        if self.converter_queue is not None:
            self.converter_queue.cancel()
        if self.filelist.discoverers is not None:
            self.filelist.discoverers.cancel()
        self.widget.hide()
        self.widget.destroy()
        # wait one second…
        # yes, this sucks badly, but signals can still be called by gstreamer
        # so wait a bit for things to calm down, and quit.
        # It can be optionally changed in the settings dict to speed up tests.
        gtk_sleep(settings.get('gtk_close_sleep', 1))
        Gtk.main_quit()
        return True

    on_window_delete_event = close
    on_quit_activate = close
    on_quit_button_clicked = close

    def on_add_activate(self, *args):
        last_folder = self.prefs.settings.get_string('last-used-folder')
        if last_folder:
            self.addchooser.set_current_folder_uri(last_folder)

        ret = self.addchooser.run()
        folder = self.addchooser.get_current_folder_uri()
        self.addchooser.hide()
        if ret == Gtk.ResponseType.OK and folder:
            self.filelist.add_uris(self.addchooser.get_uris())
            self.prefs.settings.set_string('last-used-folder', folder)
        self.set_sensitive()

    def addfile_filter_cb(self, info, pattern):
        filename = info.display_name
        return filename.lower().endswith(pattern[1:])

    def on_addfile_combo_changed(self, w):
        """Set a new filter for the filechooserwidget."""
        filefilter = Gtk.FileFilter()
        if self.addfile_combo.get_active():
            filefilter.add_custom(
                Gtk.FileFilterFlags.DISPLAY_NAME,
                self.addfile_filter_cb,
                self.pattern[self.addfile_combo.get_active()]
            )
        else:
            filefilter.add_pattern('*.*')
        self.addchooser.set_filter(filefilter)

    def on_addfolder_activate(self, *args):
        last_folder = self.prefs.settings.get_string('last-used-folder')
        if last_folder:
            self.addfolderchooser.set_current_folder_uri(last_folder)

        ret = self.addfolderchooser.run()
        folders = self.addfolderchooser.get_uris()
        folder = self.addfolderchooser.get_current_folder_uri()
        self.addfolderchooser.hide()
        if ret == Gtk.ResponseType.OK:
            extensions = None
            if self.combo.get_active():
                patterns = filepattern[self.combo.get_active()][1].split(';')
                extensions = [os.path.splitext(p)[1] for p in patterns]
            self.filelist.add_uris(folders, None, extensions)
            if folder:
                self.prefs.settings.set_string('last-used-folder', folder)

        self.set_sensitive()

    def on_remove_activate(self, *args):
        model, paths = self.filelist_selection.get_selected_rows()
        while paths:
            # Remove files
            childpath = model.convert_path_to_child_path(paths[0])
            i = self.filelist.model.get_iter(childpath)
            self.filelist.remove(i)
            model, paths = self.filelist_selection.get_selected_rows()
        # re-assign row numbers
        files = self.filelist.get_files()
        for i, sound_file in enumerate(files):
            sound_file.filelist_row = i
        self.set_sensitive()

    def on_clearlist_activate(self, *args):
        self.filelist.model.clear()
        self.filelist.filelist.clear()
        self.filelist.invalid_files_list = []
        self.invalid_files_button.set_visible(False)
        self.set_sensitive()
        self.set_status()

    def on_showinvalid_activate(self, *args):
        self.showinvalid_dialog_label.set_label(
            'Those are the files that could '
            'not be added to the list due to not\ncontaining audio data, '
            'being broken or being incompatible to gstreamer:'
        )
        buffer = Gtk.TextBuffer()
        buffer.set_text('\n'.join(self.filelist.invalid_files_list))
        self.showinvalid_dialog_list.set_buffer(buffer)
        self.showinvalid_dialog.run()
        self.showinvalid_dialog.hide()

    def do_convert(self):
        """Start the conversion."""
        name_generator = TargetNameGenerator()
        files = self.filelist.get_files()
        self.converter_queue = TaskQueue()
        self.converter_queue.connect('done', self.on_queue_finished)
        for sound_file in files:
            gtk_iteration()
            self.converter_queue.add(Converter(
                sound_file,
                name_generator
            ))
        # all was OK
        self.set_status()
        self.converter_queue.run()

        # try to make the progress bars look smooth by calling this often
        self.update_progress()
        GLib.timeout_add(1000 / 20, self.update_progress)

        # since the remining time shows only seconds, there is no need to
        # call it more often than once per second
        self.update_remaining()
        GLib.timeout_add(1000, self.update_remaining)

        self.set_sensitive()

    def update_remaining(self):
        """Refresh the remaining time in the title bar and bottom left.

        Can be used in GLib.timeout_add.
        """
        paused = self.converter_queue.paused
        running = len(self.converter_queue.running) > 0

        if not running:
            # conversion done
            self.filelist.hide_row_progress()
            return False

        if not paused and running:
            converter_queue = self.converter_queue

            if converter_queue is None:
                self.progressfile.set_markup('')
                self.filelist.hide_row_progress()
                self.progressbar.set_show_text(False)
                return

            if converter_queue.paused:
                self.progressbar.set_text(_('Paused'))
                title = '{} - {}'.format(_('SoundConverter'), _('Paused'))
                self.widget.set_title(title)
                return

            # how long it has already been running
            duration = converter_queue.get_duration()
            if duration < 1:
                # wait a bit not to display crap
                self.progressbar.set_text(_('Estimating…'))
                self.progressbar.set_show_text(True)
                return

            # remainign duration
            remaining = converter_queue.get_remaining()
            if remaining is not None:
                seconds = max(remaining % 60, 1)
                minutes = remaining / 60
                remaining = _('%d:%02d left') % (minutes, seconds)
                self.progressbar.set_text(remaining)
                self.progressbar.set_show_text(True)
                title = '{} - {}'.format(_('SoundConverter'), remaining)
                self.widget.set_title(title)

        # return True to keep the GLib timeout running
        return True

    def update_progress(self):
        """Refresh all progress bars, including the total progress.

        Can be used in GLib.timeout_add.
        """
        paused = self.converter_queue.paused
        running = len(self.converter_queue.running) > 0

        if not running:
            # conversion done
            self.filelist.hide_row_progress()
            return False

        self.filelist.show_row_progress()

        if not paused and running:
            # if paused, don't refresh the progress
            total_progress, task_progress = self.converter_queue.get_progress()
            self.progressbar.set_fraction(total_progress)

            for task, progress in task_progress:
                if progress == 0:
                    # otherwise the ui becomes really laggy with too many files
                    continue

                self.set_file_progress(task.sound_file, progress)

        # return True to keep the GLib timeout running
        return True

    def on_convert_button_clicked(self, *args):
        # reset and show progress bar
        self.progress_frame.show()
        self.status_frame.hide()
        self.set_status(_('Converting'))
        for soundfile in self.filelist.get_files():
            self.set_file_progress(soundfile, 0.0)
        # start conversion
        self.do_convert()
        # update ui
        self.set_sensitive()

    def on_button_pause_clicked(self, *args):
        if self.converter_queue.paused:
            self.converter_queue.resume()
        else:
            self.converter_queue.pause()

    def on_button_cancel_clicked(self, *args):
        self.converter_queue.cancel()
        self.set_status(_('Canceled'))
        self.set_sensitive()
        self.conversion_ended()

    def on_select_all_activate(self, *args):
        self.filelist.widget.get_selection().select_all()

    def on_clear_activate(self, *args):
        self.filelist.widget.get_selection().unselect_all()

    def on_preferences_activate(self, *args):
        self.prefs.run()

    on_prefs_button_clicked = on_preferences_activate

    def on_about_activate(self, *args):
        about = self.aboutdialog
        about.set_property('name', NAME)
        about.set_property('version', VERSION)
        about.set_transient_for(self.widget)
        # TODO: about.set_property('translator_credits', TRANSLATORS)
        about.run()

    def on_aboutdialog_response(self, *args):
        self.aboutdialog.hide()

    def selection_changed(self, *args):
        self.set_sensitive()

    def on_queue_finished(self, __=None):
        """Should be called when all conversions are completed."""
        total_time = self.converter_queue.get_duration()
        msg = _('Conversion done in %s') % format_time(total_time)
        error_count = len([
            task for task in self.converter_queue.done
            if task.error
        ])
        if error_count > 0:
            msg += ', {} error(s)'.format(error_count)

        logger.info(msg)

        self.conversion_ended(msg)

    def conversion_ended(self, msg=None):
        """Reset the window.

        Parameters
        ----------
        msg : string
            If set, will display this on the bottom left.
        """
        self.progress_frame.hide()
        self.filelist.hide_row_progress()
        self.status_frame.show()
        self.widget.set_sensitive(True)
        self.set_status(msg)
        try:
            from gi.repository import Unity
            name = "soundconverter.desktop"
            launcher = Unity.LauncherEntry.get_for_desktop_id(name)
            launcher.set_property("progress_visible", False)
        except ImportError:
            pass

    def set_widget_sensitive(self, name, sensitivity):
        self.sensitive_widgets[name].set_sensitive(sensitivity)

    def is_running(self):
        """Is a conversion (both paused and running) currently going on?"""
        queue = self.converter_queue
        return queue is not None and queue.running

    def set_sensitive(self):
        """Update the sensitive state of UI for the current state."""
        for widget_name in self.unsensitive_when_converting:
            self.set_widget_sensitive(widget_name, not self.is_running())

        if not self.is_running():
            self.set_widget_sensitive(
                'remove',
                self.filelist_selection.count_selected_rows() > 0
            )
            self.set_widget_sensitive(
                'convert_button',
                self.filelist.is_nonempty()
            )

    def set_file_progress(self, sound_file, progress):
        """Show the progress bar of a single file in the UI."""
        row = sound_file.filelist_row
        self.filelist.set_row_progress(row, progress)

    def set_status(self, text=None, ready=True):
        if not text:
            text = _('Ready')
        if ready:
            self.widget.set_title(_('SoundConverter'))
        self.statustext.set_markup(text)
        self.set_sensitive()
        gtk_iteration(True)

    def is_active(self):
        return self.widget.is_active()


NAME = VERSION = None
# use a global array as pointer, so that the constructed
# SoundConverterWindow can be accessed from unittests
win = [None]

print(_('Paused'))
import gettext
print(gettext.gettext('Ready'))
raise SystemExit

class ErrorDialog:
    def __init__(self, builder):
        self.dialog = builder.get_object('error_dialog')
        self.dialog.set_transient_for(builder.get_object('window'))
        self.primary = builder.get_object('primary_error_label')
        self.secondary = builder.get_object('secondary_error_label')

    def show_error(self, primary, secondary=None):
        self.primary.set_markup(str(primary))
        self.secondary.set_markup(str(secondary) if secondary else '')
        if secondary:
            logger.error('{}: {}'.format(primary, secondary))
        else:
            logger.error(primary)
        self.dialog.run()
        self.dialog.hide()


def gui_main(name, version, gladefile, input_files):
    """Launch the soundconverter in GTK GUI mode.

    The values for name, version and gladefile are
    determined during `make` and provided when this
    function is called in soundconverter.py

    input_files is an array of string paths, read from
    the command line arguments. It can also be an empty
    array since the user interface provides the tools
    for adding files.
    """
    global NAME, VERSION
    NAME, VERSION = name, version
    GLib.set_application_name(name)
    GLib.set_prgname(name)

    input_files = list(map(filename_to_uri, input_files))

    builder = Gtk.Builder()
    builder.set_translation_domain(name.lower())
    builder.add_from_file(gladefile)

    window = SoundConverterWindow(builder)

    set_error_handler(ErrorDialog(builder))

    window.filelist.add_uris(input_files)
    window.set_sensitive()

    global win
    win[0] = window

    Gtk.main()
